"""
Make plotly plots for html output by dials.scale, dials.report or xia2.report.
"""

from __future__ import annotations

import itertools
import math

import numpy as np
from scipy.stats import norm

from dxtbx import flumpy
from scitbx import math as scitbxmath

from dials.array_family import flex
from dials_scaling_ext import calc_lookup_index, calc_theta_phi


def _get_smooth_plotting_data_from_model(model, component="scale"):
    """Return a tuple of phis, parameters, parameter esds,
    sample positions for plotting and sample scale values."""
    valid_osc = model.configdict["valid_osc_range"]
    if model.id_ == "physical":
        sample_values = flex.double(
            np.linspace(
                valid_osc[0],
                valid_osc[1],
                int((valid_osc[1] - valid_osc[0]) / 0.1) + 1,
                endpoint=True,
            )
        )  # Make a grid of
        # points with 10 points per degree.
    elif model.id_ == "dose_decay":
        sample_values = flex.double(
            np.linspace(
                0.0,
                abs(valid_osc[1] - valid_osc[0]),
                int((abs(valid_osc[1] - valid_osc[0])) / 0.1) + 1,
                endpoint=True,
            )
        )
    if component == "scale":
        if "scale" in model.configdict["corrections"]:
            scale_SF = model.components["scale"]
            norm_vals = sample_values / model.configdict["scale_rot_interval"]
            scale_SF.data = {"x": norm_vals}
            scale_SF.update_reflection_data()
            s = scale_SF.calculate_scales()
            offset = valid_osc[0]
            if model.id_ == "dose_decay":
                offset = 0
            smoother_phis = [
                (i * model.configdict["scale_rot_interval"]) + offset
                for i in scale_SF.smoother.positions()
            ]
        return (
            smoother_phis,
            scale_SF.parameters,
            scale_SF.parameter_esds,
            sample_values,
            s,
        )
    if component == "decay":
        if "decay" in model.configdict["corrections"]:
            decay_SF = model.components["decay"]
            norm_vals = sample_values / model.configdict["decay_rot_interval"]
            decay_SF.data = {"x": norm_vals, "d": flex.double(norm_vals.size(), 1.0)}
            decay_SF.update_reflection_data()
            s = decay_SF.calculate_scales()
            smoother_phis = [
                (i * model.configdict["decay_rot_interval"]) + valid_osc[0]
                for i in decay_SF._smoother.positions()
            ]
        return (
            smoother_phis,
            decay_SF.parameters,
            decay_SF.parameter_esds,
            sample_values,
            s,
        )


smooth_help_msg = """
The inverse scale factor g, for a given reflection i, is defined as the product
of the individual model components.
For the physical scaling model, up to three terms are defined:
smoothly-varying scaling term Ci,
isotropic radiation damage term Ri = exp( Bi / (2 * di * di) ),
absorption surface correction Si,
The scaled intensity is therefore given by Ii / gi, where
gi = Ci * Ri * Si
"""


def plot_relative_Bs(relative_Bs):
    """Make scatter plot of relative Bs for dose-decay"""
    d = {
        "dose_relative_Bs": {
            "data": [
                {
                    "x": list(range(0, len(relative_Bs))),
                    "y": relative_Bs,
                    "type": "scatter",
                    "name": "Relative B-factor per dataset",
                    "mode": "markers",
                }
            ],
            "layout": {
                "title": "Relative B-factor per dataset",
                "xaxis": {"anchor": "y", "title": "Sweep id"},
                "yaxis": {"anchor": "x", "title": "Relative B (Angstrom^2)"},
            },
        }
    }
    return d


def plot_dose_decay(dose_decay_model):
    """Plot the decay and scale corrections for the dose-decay model."""
    d = {
        "dose_decay": {
            "data": [],
            "layout": {
                "title": "Dose-decay model corrections",
                "xaxis": {
                    "domain": [0, 1],
                    "anchor": "y",
                    "title": "phi (degrees)",
                },
                "yaxis": {
                    "domain": [0, 0.45],
                    "anchor": "x",
                    "title": "Decay correction <br>scale factor",
                },
                "yaxis2": {
                    "domain": [0.50, 0.95],
                    "anchor": "x",
                    "title": "Scale correction <br>scale factor",
                },
            },
        }
    }

    data = []

    if "scale" in dose_decay_model.components:
        data = _add_smooth_scales_to_data(dose_decay_model, data, yaxis="y2")

    if "decay" in dose_decay_model.components:
        data = _add_decay_model_scales_to_data(
            dose_decay_model,
            data,
            yaxis="y",
            resolution=3.0,
        )

    d["dose_decay"]["data"] = data

    return d


def _add_smooth_scales_to_data(physical_model, data, yaxis="y2"):
    (
        smoother_phis,
        parameters,
        parameter_esds,
        sample_values,
        sample_scales,
    ) = _get_smooth_plotting_data_from_model(physical_model, component="scale")

    data.append(
        {
            "x": list(sample_values),
            "y": list(sample_scales),
            "type": "line",
            "name": "smoothly-varying <br>scale correction",
            "xaxis": "x",
            "yaxis": yaxis,
        }
    )
    data.append(
        {
            "x": list(smoother_phis),
            "y": list(parameters),
            "type": "scatter",
            "mode": "markers",
            "name": "smoothly-varying <br>scale parameters",
            "xaxis": "x",
            "yaxis": yaxis,
        }
    )
    if parameter_esds:
        data[-1]["error_y"] = {"type": "data", "array": list(parameter_esds)}
    return data


def _add_decay_model_scales_to_data(model, data, yaxis="y", resolution=3.0):
    configdict = model.configdict
    valid_osc = configdict["valid_osc_range"]

    if model.id_ == "physical":
        sample_values = flex.double(
            np.linspace(
                valid_osc[0],
                valid_osc[1],
                int((valid_osc[1] - valid_osc[0]) / 0.1) + 1,
                endpoint=True,
            )
        )  # Make a grid of
        # points with 10 points per degree.
    elif model.id_ == "dose_decay":
        sample_values = flex.double(
            np.linspace(
                0.0,
                abs(valid_osc[1] - valid_osc[0]),
                int((abs(valid_osc[1] - valid_osc[0])) / 0.1) + 1,
                endpoint=True,
            )
        )
    d = flex.double(sample_values.size(), resolution)

    if "decay" in model.components:
        decay_SF = model.components["decay"]
        decay_SF.data = {"x": sample_values, "d": d}
        decay_SF.update_reflection_data()
        s = decay_SF.calculate_scales()
    data.append(
        {
            "x": list(sample_values),
            "y": list(s),
            "type": "line",
            "name": f"Decay scale factor <br>at {resolution} Angstrom",
            "xaxis": "x",
            "yaxis": yaxis,
        }
    )
    return data


def plot_smooth_scales(physical_model):
    """Plot smooth scale factors for the physical model."""

    d = {
        "smooth_scale_model": {
            "data": [],
            "layout": {
                "title": "Smoothly varying corrections",
                "xaxis": {
                    "domain": [0, 1],
                    "anchor": "y",
                    "title": "rotation angle (degrees)",
                },
                "yaxis": {
                    "domain": [0.0, 0.30],
                    "anchor": "x",
                    "title": "scale factor",
                },
                "yaxis2": {
                    "domain": [0.35, 0.65],
                    "anchor": "x",
                    "title": "relative <br> B-factor",
                },
                "yaxis3": {
                    "domain": [0.70, 1.0],
                    "anchor": "x",
                    "title": "scale factor",
                },
            },
            "help": smooth_help_msg,
        }
    }

    data = []

    if "scale" in physical_model.components:
        data = _add_smooth_scales_to_data(physical_model, data, yaxis="y3")
    if "decay" in physical_model.components:
        (
            smoother_phis,
            parameters,
            parameter_esds,
            sample_values,
            sample_scales,
        ) = _get_smooth_plotting_data_from_model(physical_model, component="decay")

        data.append(
            {
                "x": list(sample_values),
                "y": list(np.log(sample_scales) * 2.0),
                "type": "line",
                "name": "smoothly-varying <br>B-factor correction",
                "xaxis": "x",
                "yaxis": "y2",
            }
        )
        data.append(
            {
                "x": list(smoother_phis),
                "y": list(parameters),
                "type": "scatter",
                "mode": "markers",
                "name": "smoothly-varying <br>B-factor parameters",
                "xaxis": "x",
                "yaxis": "y2",
            }
        )
        if parameter_esds:
            data[-1]["error_y"] = {"type": "data", "array": list(parameter_esds)}

        data = _add_decay_model_scales_to_data(
            physical_model, data, yaxis="y", resolution=3.0
        )
    d["smooth_scale_model"]["data"].extend(data)
    return d


absorption_help_msg = """
This plot shows the smoothly-varying absorption surface used to correct the
data for the effects of absorption. It is important to note that this plot does
not show the correction applied; the applied correction for a given reflection
with scattering vectors s0, s1 is given by the average of the two values on this
surface where the surface intersects those scattering vectors. The plot is in the
crystal reference frame, and the pole (polar angle 0) corresponds to the laboratory
x-axis.

The absorption correction uses a set of spherical harmonic functions as the
basis of a smoothly varying absorption correction as a function of phi and
theta (relative to the crystal reference frame). The correction is given by:
S(s0, s1) = 1 + sum [Clm * (Ylm(s1) + Ylm(s0)) /2]
where each Ylm is a spherical harmonic and each Clm is a model parameter.
The parameters (Clm values) are shown in the 'Absorption correction surface
parameters' plot. For each l=1,2,...,lmax, there are 2l+1 m-values
from -l, -l+1,...,l+1,l, shown in the plot from left to right for each l.
s0 and s1 are the incoming and scattered beam vectors.
"""


def plot_absorption_parameters(physical_model):
    """Make a simple plot showing the absorption parameters and errors."""
    params = physical_model.components["absorption"].parameters
    param_esds = physical_model.components["absorption"].parameter_esds
    d = {
        "absorption_parameters": {
            "data": [
                {
                    "x": [i + 0.5 for i in range(len(params))],
                    "y": list(params),
                    "type": "scatter",
                    "name": "absorption parameters",
                    "xaxis": "x",
                    "yaxis": "y",
                    "mode": "markers",
                }
            ],
            "layout": {
                "title": "Absorption correction surface parameters",
                "xaxis": {"domain": [0, 1], "anchor": "y", "title": "", "tickvals": []},
                "yaxis": {"domain": [0, 1], "anchor": "x", "title": "Parameter value"},
            },
            "help": absorption_help_msg,
        }
    }
    if param_esds:
        d["absorption_parameters"]["data"][-1]["error_y"] = {
            "type": "data",
            "array": list(param_esds),
        }

    light_grey = "#d3d3d3"
    grey = "#808080"
    shapes = []
    lmax = int(-1 + (1 + len(params)) ** 0.5)
    ls = [i + 1 for i in range(lmax)]
    ns = [(2 * l) + 1 for l in ls]
    annotations = []
    start = 0
    for i, n in enumerate(ns):
        fillcolor = [light_grey, grey][i % 2]  # alternate colours
        shapes.append(
            {
                "type": "rect",
                "xref": "x",
                "yref": "paper",
                "x0": start,
                "y0": 0,
                "x1": start + n,
                "y1": 1,
                "fillcolor": fillcolor,
                "opacity": 0.2,
                "line": {"width": 0},
            }
        )
        annotations.append(
            {
                "xref": "x",
                "yref": "paper",
                "x": start + (n / 2.0),
                "y": 1,
                "text": f"l={ls[i]}",
                "showarrow": False,
                "yshift": 20,
            }
        )
        start += n
    d["absorption_parameters"]["layout"]["shapes"] = shapes
    d["absorption_parameters"]["layout"]["annotations"] = annotations
    return d


def plot_absorption_plots(physical_model, reflection_table=None):
    """Make a number of plots to help with the interpretation of the
    absorption correction."""
    # First plot the absorption surface

    d = {
        "absorption_surface": {
            "data": [],
            "layout": {
                "title": "Absorption correction surface",
                "xaxis": {
                    "domain": [0, 1],
                    "anchor": "y",
                    "title": "azimuthal angle (degrees)",
                },
                "yaxis": {
                    "domain": [0, 1],
                    "anchor": "x",
                    "title": "polar angle (degrees)",
                },
            },
            "help": absorption_help_msg,
        }
    }

    params = physical_model.components["absorption"].parameters

    order = int(-1.0 + ((1.0 + len(params)) ** 0.5))
    lfg = scitbxmath.log_factorial_generator(2 * order + 1)
    STEPS = 50
    azimuth_ = np.linspace(0, 2 * np.pi, 2 * STEPS)
    polar_ = np.linspace(0, np.pi, STEPS)
    THETA, _ = np.meshgrid(azimuth_, polar_, indexing="ij")
    lmax = int(-1.0 + ((1.0 + len(params)) ** 0.5))
    Intensity = np.ones(THETA.shape)
    undiffracted_intensity = np.ones(THETA.shape)
    counter = 0
    sqrt2 = math.sqrt(2)
    nsssphe = scitbxmath.nss_spherical_harmonics(order, 50000, lfg)
    for l in range(1, lmax + 1):
        for m in range(-l, l + 1):
            for it, t in enumerate(polar_):
                for ip, p in enumerate(azimuth_):
                    Ylm = nsssphe.spherical_harmonic(l, abs(m), t, p)
                    if m < 0:
                        r = sqrt2 * ((-1) ** m) * Ylm.imag
                    elif m == 0:
                        assert Ylm.imag == 0.0
                        r = Ylm.real
                    else:
                        r = sqrt2 * ((-1) ** m) * Ylm.real
                    Intensity[ip, it] += params[counter] * r
                    # for the undiffracted intensity, we want to add the correction
                    # at each point to the parity conjugate. We can use the fact
                    # that the odd l terms are parity odd, and even are even, to
                    # just calculate the even terms as follows
                    if l % 2 == 0:
                        undiffracted_intensity[ip, it] += params[counter] * r
            counter += 1
    d["absorption_surface"]["data"].append(
        {
            "x": list(azimuth_ * 180.0 / np.pi),
            "y": list(polar_ * 180.0 / np.pi),
            "z": list(Intensity.T.tolist()),
            "type": "heatmap",
            "colorscale": "Viridis",
            "colorbar": {"title": "inverse <br>scale factor"},
            "name": "absorption surface",
            "xaxis": "x",
            "yaxis": "y",
        }
    )

    d["undiffracted_absorption_surface"] = {
        "data": [],
        "layout": {
            "title": "Undiffracted absorption correction",
            "xaxis": {
                "domain": [0, 1],
                "anchor": "y",
                "title": "azimuthal angle (degrees)",
            },
            "yaxis": {
                "domain": [0, 1],
                "anchor": "x",
                "title": "polar angle (degrees)",
            },
        },
        "help": """
This plot shows the calculated relative absorption for a paths travelling
straight through the crystal at a given direction in a crystal-fixed frame of
reference (in spherical coordinates). This gives an indication of the effective
shape of the crystal for absorbing x-rays. In this plot, the pole (polar angle 0)
corresponds to the laboratory x-axis.
""",
    }

    d["undiffracted_absorption_surface"]["data"].append(
        {
            "x": list(azimuth_ * 180.0 / np.pi),
            "y": list(polar_ * 180.0 / np.pi),
            "z": list(undiffracted_intensity.T.tolist()),
            "type": "heatmap",
            "colorscale": "Viridis",
            "colorbar": {"title": "inverse <br>scale factor"},
            "name": "Undiffracted absorption correction",
            "xaxis": "x",
            "yaxis": "y",
        }
    )

    if not reflection_table:
        return d

    # now plot the directions of the scattering vectors

    d["vector_directions"] = {
        "data": [],
        "layout": {
            "title": "Scattering vectors in crystal frame",
            "xaxis": {
                "domain": [0, 1],
                "anchor": "y",
                "title": "azimuthal angle (degrees)",
                "range": [0, 360],
            },
            "yaxis": {
                "domain": [0, 1],
                "anchor": "x",
                "title": "polar angle (degrees)",
                "range": [0, 180],
            },
            "coloraxis": {
                "showscale": False,
            },
        },
        "help": """
This plot shows the scattering vector directions in the crystal reference frame
used to determine the absorption correction. The s0 vectors are plotted in yellow,
the s1 vectors are plotted in teal. This gives an indication of which parts of
the absorption correction surface are sampled when determining the absorption
correction. In this plot, the pole (polar angle 0) corresponds to the laboratory
x-axis.""",
    }

    STEPS = 180  # do one point per degree
    azimuth_ = np.linspace(0, 2 * np.pi, 2 * STEPS)
    polar_ = np.linspace(0, np.pi, STEPS)
    THETA, _ = np.meshgrid(azimuth_, polar_, indexing="ij")
    Intensity = np.full(THETA.shape, np.NAN)

    # note, the s1_lookup, s0_lookup is only calculated for large datasets, so
    # for small datasets we need to calculate again.
    if "s1_lookup" not in physical_model.components["absorption"].data:
        s1_lookup = calc_lookup_index(
            calc_theta_phi(reflection_table["s1c"]), points_per_degree=1
        )
        idx_polar, idx_azimuth = np.divmod(np.unique(s1_lookup), 360)
        Intensity[idx_azimuth, idx_polar] = 1
    else:
        s1_lookup = np.unique(physical_model.components["absorption"].data["s1_lookup"])
        # x is phi, y is theta
        idx_polar, idx_azimuth = np.divmod(s1_lookup, 720)
        idx_polar = idx_polar // 2  # convert from two points per degree to one
        idx_azimuth = idx_azimuth // 2
        Intensity[idx_azimuth, idx_polar] = 1

    d["vector_directions"]["data"].append(
        {
            "x": list(azimuth_ * 180.0 / np.pi),
            "y": list(polar_ * 180.0 / np.pi),
            "z": list(Intensity.T.tolist()),
            "type": "heatmap",
            "colorscale": "Viridis",
            "showscale": False,
            "xaxis": "x",
            "yaxis": "y",
            "zmin": 0,
            "zmax": 2,
        }
    )

    Intensity = np.full(THETA.shape, np.NAN)

    if "s0_lookup" not in physical_model.components["absorption"].data:
        s0_lookup = calc_lookup_index(
            calc_theta_phi(reflection_table["s0c"]), points_per_degree=1
        )
        idx_polar, idx_azimuth = np.divmod(np.unique(s0_lookup), 360)
        Intensity[idx_azimuth, idx_polar] = 2
    else:
        s0_lookup = np.unique(physical_model.components["absorption"].data["s0_lookup"])
        # x is phi, y is theta
        idx_polar, idx_azimuth = np.divmod(s0_lookup, 720)
        idx_polar = idx_polar // 2  # convert from two points per degree to one
        idx_azimuth = idx_azimuth // 2
        Intensity[idx_azimuth, idx_polar] = 2

    d["vector_directions"]["data"].append(
        {
            "x": list(azimuth_ * 180.0 / np.pi),
            "y": list(polar_ * 180.0 / np.pi),
            "z": list(Intensity.T.tolist()),
            "type": "heatmap",
            "colorscale": "Viridis",
            "showscale": False,
            "xaxis": "x",
            "yaxis": "y",
            "zmin": 0,
            "zmax": 2,
        }
    )

    scales = physical_model.components["absorption"].calculate_scales()
    hist = flex.histogram(scales, n_slots=min(100, int(scales.size() * 10)))

    d["absorption_corrections"] = {
        "data": [
            {
                "x": list(hist.slot_centers()),
                "y": list(hist.slots()),
                "type": "bar",
                "name": "Applied absorption corrections",
            },
        ],
        "layout": {
            "title": "Applied absorption corrections",
            "xaxis": {"anchor": "y", "title": "Inverse scale factor"},
            "yaxis": {"anchor": "x", "title": "Number of reflections"},
        },
    }

    return d


def plot_outliers(data):
    """plots positions of outliers"""

    if not data["z"]:
        return {"outlier_xy_positions": {}, "outliers_vs_z": {}}

    hist = flex.histogram(
        flex.double(data["z"]), n_slots=min(100, int(len(data["z"]) * 10))
    )

    d = {
        "outlier_xy_positions": {
            "data": [
                {
                    "x": data["x"],
                    "y": data["y"],
                    "type": "scatter",
                    "mode": "markers",
                    "xaxis": "x",
                    "yaxis": "y",
                }
            ],
            "layout": {
                "title": "Outlier x-y positions",
                "xaxis": {
                    "anchor": "y",
                    "title": "x (px)",
                    "range": [0, data["image_size"][0]],
                },
                "yaxis": {
                    "anchor": "x",
                    "title": "y (px)",
                    "range": [0, data["image_size"][1]],
                },
            },
        },
        "outliers_vs_z": {
            "data": [
                {
                    "x": list(hist.slot_centers()),
                    "y": list(hist.slots()),
                    "type": "bar",
                    "name": "outliers vs rotation",
                }
            ],
            "layout": {
                "title": "Outlier distribution across frames",
                "xaxis": {"title": "frame"},
                "yaxis": {"title": "count"},
                "bargap": 0,
            },
        },
    }

    return d


def error_model_variance_plot(data, label=None):
    bin_variances = data["binning_info"]["bin_variances"]
    initial_variances = data["binning_info"]["initial_variances"]
    xs = data["binning_info"]["bin_boundaries"]
    x = list(range(1, len(xs)))
    x_labels = [
        str(round(xs[i], 1)) + " - " + str(round(xs[i + 1], 1))
        for i in range(len(xs) - 1)
    ]
    key = (
        f"error_model_variances_{label}"
        if label is not None
        else "error_model_variances"
    )
    title = "Error model variances of normalised deviations"
    title = title + f" (error model {label})" if label is not None else title
    d = {
        key: {
            "data": [
                {
                    "x": x,
                    "y": list(initial_variances)[::-1],
                    "type": "scatter",
                    "mode": "markers",
                    "xaxis": "x",
                    "yaxis": "y",
                    "name": "Uncorrected variances",
                },
                {
                    "x": x,
                    "y": list(bin_variances)[::-1],
                    "type": "scatter",
                    "mode": "markers",
                    "xaxis": "x",
                    "yaxis": "y",
                    "name": "Corrected variances",
                },
                {
                    "x": [1, 10],
                    "y": [1, 1],
                    "type": "line",
                    "name": "Ideal normal distribution",
                },
            ],
            "layout": {
                "title": title,
                "xaxis": {
                    "anchor": "y",
                    "title": "Intensity range, expected unscaled intensity (counts)",
                    "tickvals": x,
                    "ticktext": x_labels[::-1],
                },
                "yaxis": {"anchor": "x", "title": "Variance of normalised deviations"},
            },
            "help": """\
This plot shows the variance of the normalised deviations for a given intensity
range for the error model minimisation. The expectation is that after a successful
error model correction, the variance will be one across the intensity range
measured i.e. the intensities are normally distributed about the symmetry group
best estimate, with a variance of one sigma.
""",
        }
    }
    return d


def error_regression_plot(data, label=None):
    """Plot the data from the regression fit."""
    x = data["regression_x"]
    y = data["regression_y"]
    fit = (x * (data["model_a"] ** 2)) + ((data["model_a"] * data["model_b"]) ** 2)
    key = f"regression_fit_{label}" if label is not None else "regression_fit"
    title = "Error model regression plot"
    title = title + f" (error model {label})" if label is not None else title
    return {
        key: {
            "data": [
                {
                    "x": list(x),
                    "y": list(y),
                    "type": "scatter",
                    "mode": "markers",
                    "name": "expected vs observed",
                },
                {
                    "x": list(x),
                    "y": list(fit),
                    "type": "scatter",
                    "name": "best least-squares fit",
                },
            ],
            "layout": {
                "title": title,
                "xaxis": {"anchor": "y", "title": "1/(I/sigma_obs) ^ 2"},
                "yaxis": {"anchor": "x", "title": "1/(I/sigma) ^ 2 "},
            },
            "help": """\
This plot shows the data used for the regression fit for error model refinement.
The form derives from the equation sigma^2(obs) = a^2[sigma^2 + (bI)^2].
Dividing by I^2 gives 1/(I/sigma_obs) ^ 2 = [a^2 * 1/(I/sigma) ^ 2] + [a^2 b^2]
i.e. y = mx + c.
Here, sigma_obs is the observed standard deviation of a group of symmetry
equivalents i.e. sigma_obs^2 = (Sum (I - g<Ih>)^2) / N-1.
""",
        }
    }


def normal_probability_plot(data, label=None):
    """Plot the distribution of normal probabilities of errors."""

    n = data["delta_hl"].size
    y = np.sort(data["delta_hl"])
    delta = 0.5 / n
    v = np.linspace(start=delta, stop=1.0 - delta, endpoint=True, num=n)
    x = norm.ppf(v)

    H, xedges, yedges = np.histogram2d(x, y, bins=(200, 200))
    nonzeros = np.nonzero(H)
    z = np.empty(H.shape)
    z[:] = np.NAN
    z[nonzeros] = H[nonzeros]

    # also make a histogram
    histy = flex.histogram(flumpy.from_numpy(y), n_slots=100)
    # make a gaussian for reference also
    n = y.size
    width = histy.slot_centers()[1] - histy.slot_centers()[0]
    gaussian = [
        n * width * math.exp(-(sc**2) / 2.0) / ((2.0 * math.pi) ** 0.5)
        for sc in histy.slot_centers()
    ]
    key = (
        f"normal_distribution_plot_{label}"
        if label is not None
        else "normal_distribution_plot"
    )
    title = "Normal probability plot with error model applied"
    title = title + f" (error model {label})" if label is not None else title
    key_hist = f"nor_dev_hist_{label}" if label is not None else "nor_dev_hist"
    title_hist = "Normal deviations with error model applied"
    title_hist = (
        title_hist + f" (error model {label})" if label is not None else title_hist
    )
    return {
        key: {
            "data": [
                {
                    "x": xedges.tolist(),
                    "y": yedges.tolist(),
                    "z": z.transpose().tolist(),
                    "type": "heatmap",
                    "name": "normalised deviations",
                    "colorbar": {
                        "title": "Number of reflections",
                        "titleside": "right",
                    },
                    "colorscale": "Viridis",
                },
                {
                    "x": [-5, 5],
                    "y": [-5, 5],
                    "type": "scatter",
                    "mode": "lines",
                    "name": "z = m",
                    "color": "rgb(0,0,0)",
                },
            ],
            "layout": {
                "title": title,
                "xaxis": {"anchor": "y", "title": "Order statistic medians, m"},
                "yaxis": {"anchor": "x", "title": "Ordered responses, z"},
            },
            "help": """\
This plot shows the normalised devations (of each reflection from the
group-weighted mean), sorted in order and plotted against the expected order
based on a normal distribution model. A true normal distribution of deviations
would give the straight line indicated. If the errors are well described by
this model, the ordered responses should closely fit the straight line to
high absolute values of x (>3), where there is typically a deviation away from
the line due to wide tails of the distribution.
""",
        },
        key_hist: {
            "data": [
                {
                    "x": list(histy.slot_centers()),
                    "y": list(histy.slots()),
                    "type": "bar",
                    "name": "dataset normalised deviations",
                },
                {
                    "x": list(histy.slot_centers()),
                    "y": gaussian,
                    "type": "scatter",
                    "name": "Ideal normal distribution",
                },
            ],
            "layout": {
                "title": title_hist,
                "xaxis": {"anchor": "y", "title": "Normalised deviation"},
                "yaxis": {"anchor": "x", "title": "Number of reflections"},
            },
            "help": """\
This plot shows the distribution of normalised devations (of each reflection
from the group-weighted mean), for the reflections used to minimise the error
model. A true normal distribution is indicated.
""",
        },
    }


def _smooth_params_to_bins(n_param):
    if n_param > 4:
        return n_param - 2
    return n_param - 1


def plot_array_modulation_plot(array_model):
    modulation_comp = array_model.components["modulation"]
    configdict = array_model.configdict
    nxbins = _smooth_params_to_bins(configdict["n_x_mod_param"])
    nybins = _smooth_params_to_bins(configdict["n_y_mod_param"])
    sample_x_values = flex.double(
        np.linspace(0, nxbins, (nxbins * 10 + 1), endpoint=True)
    )
    sample_y_values = flex.double(
        np.linspace(0, nybins, (nybins * 10) + 1, endpoint=True)
    )
    x_vals = flex.double(np.tile(sample_x_values, len(sample_y_values)))
    y_vals = flex.double(np.repeat(sample_y_values, len(sample_x_values)))
    modulation_comp.data = {"x": x_vals, "y": y_vals}
    modulation_comp.update_reflection_data()
    z = modulation_comp.calculate_scales()

    xtickvals = flex.double(range(nxbins + 1))
    ytickvals = flex.double(range(nybins + 1))
    xticks = (xtickvals * configdict["x_det_bin_width"]) + configdict["xmin"]
    yticks = (ytickvals * configdict["y_det_bin_width"]) + configdict["ymin"]
    xticktext = [str(int(i)) for i in xticks]
    yticktext = [str(int(i)) for i in yticks]

    return {
        "array_modulation_plot": {
            "data": [
                {
                    "x": list(x_vals),
                    "y": list(y_vals),
                    "z": list(z),
                    "type": "heatmap",
                    "colorscale": "Viridis",
                }
            ],
            "layout": {
                "title": "Array model modulation correction",
                "xaxis": {
                    "title": "detector x-position (px)",
                    "showgrid": False,
                    "tickvals": list(xtickvals),
                    "ticktext": xticktext,
                },
                "yaxis": {
                    "title": "detector y-position (px)",
                    "anchor": "x",
                    "showgrid": False,
                    "tickvals": list(ytickvals),
                    "ticktext": yticktext,
                },
                "width": 500,
                "height": 450,
            },
        }
    }


def plot_array_absorption_plot(array_model):
    absorption_comp = array_model.components["absorption"]
    configdict = array_model.configdict
    nxbins = _smooth_params_to_bins(configdict["n_x_param"])
    nybins = _smooth_params_to_bins(configdict["n_y_param"])
    ntimebins = _smooth_params_to_bins(configdict["n_time_param"])

    sample_x_values = flex.double(np.linspace(0, nxbins, nxbins + 1, endpoint=True))
    sample_y_values = flex.double(np.linspace(0, nybins, nybins + 1, endpoint=True))
    sample_time_values = flex.double(
        np.linspace(0, ntimebins, ntimebins + 1, endpoint=True)
    )
    # the x-y values at one timepoint
    x_vals = np.tile(sample_x_values, len(sample_y_values))
    y_vals = np.repeat(sample_y_values, len(sample_x_values))
    # now tile this for n_time_vals
    x_vals = np.tile(x_vals, len(sample_time_values))
    y_vals = np.tile(y_vals, len(sample_time_values))

    z_vals = flex.double(
        np.repeat(sample_time_values, len(sample_x_values) * len(sample_y_values))
    )

    absorption_comp.data = {
        "x": flex.double(x_vals),
        "y": flex.double(y_vals),
        "z": flex.double(z_vals),
    }
    absorption_comp.update_reflection_data()
    z = absorption_comp.calculate_scales()

    xs = np.repeat(
        np.array(
            sample_time_values * configdict["time_rot_interval"]
            + configdict["valid_osc_range"][0]
        ),
        len(sample_x_values) * len(sample_y_values),
    )
    ys = list(
        itertools.chain(
            *itertools.repeat(
                range(len(sample_x_values) * len(sample_y_values)),
                len(sample_time_values),
            )
        )
    )

    return {
        "array_absorption_plot": {
            "data": [
                {
                    "x": list(xs),
                    "y": list(ys),
                    "z": list(z),
                    "type": "heatmap",
                    "colorscale": "Viridis",
                }
            ],
            "layout": {
                "title": "Array model absorption correction",
                "xaxis": {"title": "rotation angle (degrees)", "showgrid": False},
                "yaxis": {
                    "title": "detector x-y positional index",
                    "anchor": "x",
                    "showgrid": False,
                },
                "width": 500,
                "height": 450,
            },
            "help": """\
This plot shows a representation of the 3D absorption correction unwrapped into
2D. For each value of rotation (separated by ~decay_interval), the y-axis shows
the correction at the corners of a grid of n_absorption_bins x n_absorption_bins.
e.g. if n_absorption_bins = 3, there are 16 corners of the grid that spans the
detector surface, so the positional indices are the points (0, 0), (0, 1), (0, 2),
(0, 3), (1, 0), (1, 1) etc. on this grid. Although the plot is granular, the
correction applied during scaling is smoothly interpolated between the parameters
in 3D.
""",
        }
    }


def plot_array_decay_plot(array_model):

    decay_comp = array_model.components["decay"]
    configdict = array_model.configdict

    valid_osc = configdict["valid_osc_range"]
    n_points = max(int(math.ceil(valid_osc[1] - valid_osc[0])), 50)
    sample_x_values = flex.double(
        np.linspace(valid_osc[0], valid_osc[1], n_points + 1, endpoint=True)
    )
    n_y_bins = _smooth_params_to_bins(configdict["n_res_param"])
    sample_y_values = np.linspace(0, n_y_bins, int(n_y_bins / 0.1) + 1, endpoint=True)

    norm_x_vals = (
        flex.double(np.tile(sample_x_values, len(sample_y_values)))
        / configdict["time_rot_interval"]
    )
    norm_y_vals = flex.double(np.repeat(sample_y_values, len(sample_x_values)))
    decay_comp.data = {"x": norm_y_vals, "y": norm_x_vals}
    decay_comp.update_reflection_data()
    z = decay_comp.calculate_scales()
    x = norm_x_vals * configdict["time_rot_interval"]
    y = norm_y_vals

    tickvals = flex.double(range(n_y_bins + 1))
    resmin = (tickvals * configdict["res_bin_width"]) + configdict["resmin"]
    d = 1.0 / flex.sqrt((tickvals * configdict["res_bin_width"]) + resmin)
    ticktext = [f"{i:.3f}" for i in d]

    return {
        "array_decay_plot": {
            "data": [
                {
                    "x": list(x),
                    "y": list(y),
                    "z": list(z),
                    "type": "heatmap",
                    "colorscale": "Viridis",
                }
            ],
            "layout": {
                "title": "Array model decay correction",
                "xaxis": {"title": "rotation angle (degrees)", "showgrid": False},
                "yaxis": {
                    "title": "resolution/d-value (Angstrom)",
                    "anchor": "x",
                    "showgrid": False,
                    "tickvals": list(tickvals),
                    "ticktext": ticktext,
                },
                "width": 500,
                "height": 450,
            },
        }
    }



from __future__ import annotations


def centroid_px_to_mm(detector, scan, position, variance, sd_error):
    """Convenience function to calculate centroid in mm/rad from px"""

    # Get the pixel to millimeter function
    assert len(detector) == 1
    return centroid_px_to_mm_panel(detector[0], scan, position, variance, sd_error)


def centroid_px_to_mm_panel(panel, scan, position, variance, sd_error):
    """Convenience function to calculate centroid in mm/rad from px"""
    # Get the pixel to millimeter function
    pixel_size = panel.get_pixel_size()
    if scan is None:
        oscillation = (0, 0)
    else:
        oscillation = scan.get_oscillation(deg=False)
    scale = pixel_size + (oscillation[1],)
    scale2 = tuple(s * s for s in scale)

    if isinstance(position, tuple):
        # Convert Pixel coordinate into mm/rad
        x, y, z = position
        xy_mm = panel.pixel_to_millimeter((x, y))

        if scan is None:
            z_rad = 0
        else:
            z_rad = scan.get_angle_from_array_index(z, deg=False)

        # Set the position, variance and squared width in mm/rad
        # N.B assuming locally flat pixel to millimeter transform
        # for variance calculation.
        position_mm = xy_mm + (z_rad,)
        variance_mm = [var * s for var, s in zip(variance, scale2)]
        sd_error_mm = [sde * s for sde, s in zip(sd_error, scale2)]

    else:
        from scitbx.array_family import flex

        # Convert Pixel coordinate into mm/rad
        x, y, z = position.parts()
        xy_mm = panel.pixel_to_millimeter(flex.vec2_double(x, y))

        if scan is None:
            z_rad = flex.double(z.size(), 0)
        else:
            z_rad = scan.get_angle_from_array_index(z, deg=False)

        # Set the position, variance and squared width in mm/rad
        # N.B assuming locally flat pixel to millimeter transform
        # for variance calculation.
        x_mm, y_mm = xy_mm.parts()
        position_mm = flex.vec3_double(x_mm, y_mm, z_rad)
        v0, v1, v2 = variance.parts()
        variance_mm = flex.vec3_double(v0 * scale2[0], v1 * scale2[1], v2 * scale2[2])
        s0, s1, s2 = sd_error.parts()
        sd_error_mm = flex.vec3_double(s0 * scale2[0], s1 * scale2[1], s2 * scale2[2])

    # Return the stuff in mm/rad
    return position_mm, variance_mm, sd_error_mm



#!/usr/bin/python
#
# this is used by the build scripts - it should not normally need to be
# called in a full DIALS installation (although this will probably work too)
"""
Echo the current DIALS version and build number.
"""


from __future__ import annotations

import datetime

# XXX TO CHANGE THE DIALS VERSION NAME, EDIT BASE_VERSION
BASE_VERSION = "dev"
# this is when the first DIALS installer was created (UTC time)
start_date = datetime.date(2014, 7, 10)
today = datetime.datetime.utcnow().date()
delta = today - start_date
n_days = abs(delta.days) + 1
print("%s-%d" % (BASE_VERSION, n_days))



from __future__ import annotations


class ProfileModelExt:
    """
    The definition for a profile model.
    """

    @staticmethod
    def create(
        params, reflections, crystal, beam, detector, goniometer=None, scan=None
    ):
        """
        Create the profile model from data.

        :param params: The phil parameters
        :param reflections: The reflections
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        :return: An instance of the profile model
        """
        return None

    def predict_reflections(
        self, imageset, crystal, beam, detector, goniometer=None, scan=None, **kwargs
    ):
        """
        Given an experiment, predict the reflections.

        :param imageset: The imageset
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        pass

    def compute_partiality(
        self, reflections, crystal, beam, detector, goniometer=None, scan=None, **kwargs
    ):
        """
        Given an experiment and list of reflections, compute the partiality of the
        reflections

        :param reflections: The reflection table
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        pass

    def compute_bbox(
        self, reflections, crystal, beam, detector, goniometer=None, scan=None, **kwargs
    ):
        """Given an experiment and list of reflections, compute the
        bounding box of the reflections on the detector (and image frames).

        :param reflections: The reflection table
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        pass

    def compute_mask(
        self, reflections, crystal, beam, detector, goniometer=None, scan=None, **kwargs
    ):
        """
        Given an experiment and list of reflections, compute the
        foreground/background mask of the reflections.

        :param reflections: The reflection table
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        pass

    def fitting_class(self):
        """
        Get the profile fitting algorithm associated with this profile model

        :return: The profile fitting class
        """
        return None



from __future__ import annotations

import math
import random

import pytest

from scitbx import matrix

from dials.algorithms.profile_model.gaussian_rs import CoordinateSystem

### Test the XDS coordinate system class


@pytest.fixture
def xdscoordinates():
    """Initialise the coordinate system"""
    # Coordinate sensitive to length of vectors, so need to ensure that
    # lengths of both s0 and s1 are equal
    coords = {
        "s0": (0.013141995425357206, 0.002199999234194632, 1.4504754950989514),
        "s1": (-0.01752795848400313, -0.24786554213968193, 1.4290948735525306),
        "m2": (0.999975, -0.001289, -0.006968),
        "phi": 5.83575672475 * math.pi / 180,
    }
    coords["cs"] = CoordinateSystem(
        coords["m2"], coords["s0"], coords["s1"], coords["phi"]
    )
    return coords


def test_coordinate_system_data(xdscoordinates):
    """Test all the input data"""
    eps = 1e-7
    s0 = matrix.col(xdscoordinates["s0"])
    s1 = matrix.col(xdscoordinates["s1"])
    m2 = matrix.col(xdscoordinates["m2"])
    assert abs(matrix.col(xdscoordinates["cs"].s0()) - s0) <= eps
    assert abs(matrix.col(xdscoordinates["cs"].s0()) - s0) <= eps
    assert abs(matrix.col(xdscoordinates["cs"].s1()) - s1) <= eps
    assert abs(matrix.col(xdscoordinates["cs"].m2()) - m2.normalize()) <= eps
    assert abs(xdscoordinates["cs"].phi() - xdscoordinates["phi"]) <= eps


def test_ensure_axes_have_length_of_one(xdscoordinates):
    eps = 1e-7
    assert abs(matrix.col(xdscoordinates["cs"].e1_axis()).length() - 1.0) <= eps
    assert abs(matrix.col(xdscoordinates["cs"].e2_axis()).length() - 1.0) <= eps
    assert abs(matrix.col(xdscoordinates["cs"].e3_axis()).length() - 1.0) <= eps


def test_axis_orthogonal(xdscoordinates):
    """Ensure that the following are true:

    e1.s0 = 0, e1.s1 = 0
    e2.s1 = 0, e2.e1 = 0
    e3.e1 = 0, e3.p* = 0
    """
    # Get as matrices
    e1 = matrix.col(xdscoordinates["cs"].e1_axis())
    e2 = matrix.col(xdscoordinates["cs"].e2_axis())
    e3 = matrix.col(xdscoordinates["cs"].e3_axis())
    s0 = matrix.col(xdscoordinates["s0"])
    s1 = matrix.col(xdscoordinates["s1"])

    eps = 1e-7

    # Check that e1 is orthogonal to s0 and s1
    assert abs(e1.dot(s0) - 0.0) <= eps
    assert abs(e1.dot(s1) - 0.0) <= eps

    # Check that e2 is orthogonal to s1 and e1
    assert abs(e2.dot(s1) - 0.0) <= eps
    assert abs(e2.dot(e1) - 0.0) <= eps

    # Check that e3 is orthogonal to e1 and p* = s1 - s0
    assert abs(e3.dot(e1) - 0.0) <= eps
    assert abs(e3.dot(s1 - s0) - 0.0) <= eps


def test_the_coordinate_system_limits(xdscoordinates):
    """Ensure limits e1/e2 == |s1| and limit e3 == |s0 - s1|"""
    # Get the incident and diffracted beam vectors
    eps = 1e-7

    # Get the limits
    lim = xdscoordinates["cs"].limits()

    # Check the limits
    assert lim[0] == pytest.approx(-1.0, abs=eps)
    assert lim[1] == pytest.approx(1.0, abs=eps)


### Test the FromBeamVectorToXds class


@pytest.fixture
def beamvector():
    """Initialise the transform"""

    bv = {
        "s0": (0.013141995425357206, 0.002199999234194632, 1.4504754950989514),
        "s1": (-0.01752795848400313, -0.24786554213968193, 1.4290948735525306),
        "m2": (0.999975, -0.001289, -0.006968),
        "phi": 5.83575672475 * math.pi / 180,
    }
    bv["cs"] = CoordinateSystem(bv["m2"], bv["s0"], bv["s1"], bv["phi"])
    return bv


def test_beamvector_coordinate_of_s1(beamvector):
    """Ensure that the coordinate of s1 is (0, 0, 0)"""

    # Get the coordinate at s1
    s_dash = beamvector["s1"]
    c1, c2 = beamvector["cs"].from_beam_vector(s_dash)

    # Ensure that it is at the origin
    assert c1 == pytest.approx(0.0)
    assert c2 == pytest.approx(0.0)


def test_beamvector_limit(beamvector):
    """Calculate the coordinate at the limits.

    Ensure that coordinate where s1' is orthogonal to s1 is at limit.
    """
    # Get the limit of s1'
    s_dash = matrix.col(beamvector["s1"]).cross(matrix.col(beamvector["s0"]))
    s_dash = s_dash.normalize() * matrix.col(beamvector["s1"]).length()

    # Rotate arbitrarily
    s_dash = s_dash.rotate_around_origin(
        matrix.col(beamvector["s1"]), random.uniform(0, 360), deg=True
    )

    # Get the c1, c2 coordinate
    c1, c2 = beamvector["cs"].from_beam_vector(s_dash)

    # Check the point is equal to the limit in rs
    assert math.sqrt(c1**2 + c2**2) == pytest.approx(
        abs(beamvector["cs"].limits()[0])
    )


### Test the TestFromRotationAngle class


@pytest.fixture
def rotationangle():
    """Initialise the transform"""

    ra = {
        "s0": (0.013141995425357206, 0.002199999234194632, 1.4504754950989514),
        "s1": (-0.01752795848400313, -0.24786554213968193, 1.4290948735525306),
        "m2": (0.999975, -0.001289, -0.006968),
        "phi": 5.83575672475 * math.pi / 180,
    }

    ra["cs"] = CoordinateSystem(ra["m2"], ra["s0"], ra["s1"], ra["phi"])
    return ra


def test_from_rotation_angle_coordinate_of_phi(rotationangle):
    """Ensure that the coordinate of s1 is (0, 0, 0)"""

    # Get the coordinate at phi
    phi_dash = rotationangle["phi"]
    c3 = rotationangle["cs"].from_rotation_angle(phi_dash)

    # Ensure that it is at the origin
    assert c3 == pytest.approx(0.0)


def test_from_rotation_angle_e3_coordinate_approximation(rotationangle):
    # Select a random rotation from phi
    phi_dash = rotationangle["phi"] + (2.0 * random.random() - 1.0) * math.pi / 180

    # Calculate the XDS coordinate, this class uses an approximation
    # for c3 = zeta * (phi' - phi)
    c3 = rotationangle["cs"].from_rotation_angle(phi_dash)
    c3_2 = rotationangle["cs"].from_rotation_angle_fast(phi_dash)

    # Check the true and approximate value are almost equal to 4dp
    assert c3 == pytest.approx(c3_2, abs=1e-4)


### Test the ToBeamVector class


def test_to_beamvector_xds_origin(beamvector):
    """Test the beam vector at the XDS origin is equal to s1."""
    eps = 1e-7
    s_dash = beamvector["cs"].to_beam_vector((0, 0))
    assert abs(matrix.col(s_dash) - matrix.col(beamvector["s1"])) <= eps


def test_to_beamvector_far_out_coordinates(beamvector):
    """Test some large coordinates, 1 valid and the other invalid. (i.e.
    a coordinate that cannot be mapped onto the ewald sphere)."""
    eps = 1e-7

    c2 = 0

    # A large value which is still valid
    c1 = 1.0 - eps
    assert beamvector["cs"].to_beam_vector((c1, c2))

    # A large value which is raises an exception
    with pytest.raises(RuntimeError):
        c1 = 1.0 + eps
        assert beamvector["cs"].to_beam_vector((c1, c2))

    c1 = 0

    # A large value which is still valid
    c2 = 1.0 - eps
    assert beamvector["cs"].to_beam_vector((c1, c2))

    # A large value which is raises an exception
    with pytest.raises(RuntimeError):
        c2 = 1.0 + eps
        assert beamvector["cs"].to_beam_vector((c1, c2))


def test_to_beamvector_forward_and_reverse_transform(beamvector):
    """Test the forward and reverse Beam Vector -> XDS transforms Create
    a beam vector, transform it to XDS and then transform back. The new
    value should be equal to the original value."""
    eps = 1e-7

    # Set the parameters
    min_shift = -0.5
    max_shift = +0.5
    range_shift = max_shift - min_shift

    def random_shift():
        return min_shift + random.random() * range_shift

    # Loop a number of times
    num = 1000
    for i in range(num):
        # Create a beam vector
        s_dash = matrix.col(beamvector["s1"]) + matrix.col(
            (random_shift(), random_shift(), random_shift())
        )
        s_dash = s_dash.normalize() * matrix.col(beamvector["s1"]).length()

        # Calculate the XDS coordinate of the vector
        c1, c2 = beamvector["cs"].from_beam_vector(s_dash)

        # Calculate the beam vector from the XDS coordinate
        s_dash_2 = beamvector["cs"].to_beam_vector((c1, c2))

        # Check the vectors are almost equal
        assert abs(matrix.col(s_dash) - matrix.col(s_dash_2)) <= eps


### Test the TestToRotationAngle class


def test_forward_and_backward(rotationangle):
    # Set the parameters
    min_shift = -20.0 * math.pi / 180.0
    max_shift = +20.0 * math.pi / 180.0
    range_shift = max_shift - min_shift

    def random_shift():
        return min_shift + random.random() * range_shift

    # Loop a number of times
    num = 1000
    for i in range(num):

        # Create a rotation angle
        phi_dash = rotationangle["phi"] + random_shift()

        # Calculate the XDS coordinate of the vector
        c3 = rotationangle["cs"].from_rotation_angle(phi_dash)

        # Calculate the beam vector from the XDS coordinate
        phi_dash_2 = rotationangle["cs"].to_rotation_angle(c3)

        # Check the vectors are almost equal
        assert phi_dash_2 == pytest.approx(phi_dash)


def test_origin(rotationangle):
    phi_dash = rotationangle["cs"].to_rotation_angle(0.0)
    assert phi_dash == pytest.approx(rotationangle["phi"])


def test_far_out_coordinates(rotationangle):
    """Test some large coordinates, 1 valid and the other invalid. (i.e.
    a coordinate that cannot be mapped onto the ewald sphere)."""

    eps = 1e-7

    # Get the limits
    lim = rotationangle["cs"].limits()[2:]
    lim0 = max(lim)
    lim1 = min(lim)

    # Setting c2 and c3 to zero
    c3 = lim0 - eps

    # A large value which is still valid
    phi_dash = rotationangle["cs"].to_rotation_angle(c3)

    # Setting c2 and c3 to zero
    c3 = lim1 + eps

    # A large value which is still valid
    phi_dash = rotationangle["cs"].to_rotation_angle(c3)

    # A large value which is raises an exception
    with pytest.raises(RuntimeError):
        c3 = lim0 + eps
        phi_dash = rotationangle["cs"].to_rotation_angle(c3)
        print(phi_dash)

    with pytest.raises(RuntimeError):
        c3 = lim1 - eps
        phi_dash = rotationangle["cs"].to_rotation_angle(c3)


def test_e3_coordinate_approximation(rotationangle):
    # Select a random rotation from phi
    phi_dash = rotationangle["phi"] + (2.0 * random.random() - 1.0) * math.pi / 180

    # Calculate the XDS coordinate, this class uses an approximation
    # for c3 = zeta * (phi' - phi)
    c3 = rotationangle["cs"].from_rotation_angle(phi_dash)
    phi_dash_2 = rotationangle["cs"].to_rotation_angle_fast(c3)

    # Check the true and approximate value are almost equal to 4dp
    assert phi_dash == pytest.approx(phi_dash_2, abs=1e-4)



from __future__ import annotations

import itertools
import logging
import math
import os
import pathlib
from time import time

import psutil

import boost_adaptbx.boost.python
import libtbx

import dials.algorithms.integration
import dials.util
import dials.util.log
from dials.array_family import flex
from dials.model.data import make_image
from dials.util import tabulate
from dials.util.log import rehandle_cached_records
from dials.util.mp import available_cores, multi_node_parallel_map
from dials_algorithms_integration_integrator_ext import (
    Executor,
    Group,
    GroupList,
    Job,
    JobList,
    ReflectionManager,
    ReflectionManagerPerImage,
    ShoeboxProcessor,
)

try:
    import resource
except ImportError:
    # resource does not exist on non-Linux, so can't float the import
    resource = None

__all__ = [
    "Block",
    "build_processor",
    "Debug",
    "Executor",
    "Group",
    "GroupList",
    "Job",
    "job",
    "JobList",
    "Lookup",
    "MultiProcessing",
    "NullTask",
    "Parameters",
    "Processor2D",
    "Processor3D",
    "ProcessorFlat3D",
    "ProcessorSingle2D",
    "ProcessorStills",
    "ReflectionManager",
    "ReflectionManagerPerImage",
    "Shoebox",
    "ShoeboxProcessor",
    "Task",
]

logger = logging.getLogger(__name__)


def assess_available_memory(params):

    # Obtain information about system memory
    available_memory = psutil.virtual_memory().available
    available_swap = psutil.swap_memory().free

    # https://htcondor.readthedocs.io/en/latest/users-manual/services-for-jobs.html#extra-environment-variables-htcondor-sets-for-jobs
    condor_job_ad = os.environ.get("_CONDOR_JOB_AD")
    if condor_job_ad:
        try:
            job_classad = dials.util.parse_htcondor_job_classad(
                pathlib.Path(condor_job_ad)
            )
        except Exception as e:
            logger.error(
                f"Error parsing _CONDOR_JOB_AD {condor_job_ad}: {e}",
                exc_info=True,
            )
        else:
            if job_classad.memory_provisioned:
                # Convert MB to bytes
                available_memory = min(
                    available_memory, job_classad.memory_provisioned * 1024**2
                )

    available_incl_swap = available_memory + available_swap
    available_limit = available_incl_swap * params.block.max_memory_usage
    available_immediate_limit = available_memory * params.block.max_memory_usage

    # Compile a memory report
    report = [
        "Memory situation report:",
    ]

    def _report(description, numbytes):
        report.append(f"  {description:<50}:{numbytes/1e9:5.1f} GB")

    _report("Available system memory (excluding swap)", available_memory)
    _report("Available swap memory", available_swap)
    _report("Available system memory (including swap)", available_incl_swap)
    _report("Maximum memory for processing (including swap)", available_limit)
    _report(
        "Maximum memory for processing (excluding swap)",
        available_immediate_limit,
    )

    # Check if a ulimit applies
    # Note that resource may be None on non-Linux platforms.
    # We can't use psutil as platform-independent solution in this instance due to
    # https://github.com/conda-forge/psutil-feedstock/issues/47
    rlimit = getattr(resource, "RLIMIT_VMEM", getattr(resource, "RLIMIT_AS", None))
    if rlimit:
        try:
            ulimit = resource.getrlimit(rlimit)[0]
            if ulimit <= 0 or ulimit > (2**62):
                report.append("  no memory ulimit set")
            else:
                ulimit_used = psutil.Process().memory_info().rss
                _report("Memory ulimit detected", ulimit)
                _report("Memory ulimit in use", ulimit_used)
                available_memory = max(0, min(available_memory, ulimit - ulimit_used))
                available_incl_swap = max(
                    0, min(available_incl_swap, ulimit - ulimit_used)
                )
                available_immediate_limit = (
                    available_memory * params.block.max_memory_usage
                )
                _report("Available system memory (limited)", available_memory)
                _report(
                    "Available system memory (incl. swap; limited)",
                    available_incl_swap,
                )
                _report(
                    "Maximum memory for processing (exc. swap; limited)",
                    available_immediate_limit,
                )
        except Exception as e:
            logger.debug(
                "Could not obtain ulimit values due to %s", str(e), exc_info=True
            )

    return available_immediate_limit, available_incl_swap, report


def _average_bbox_size(reflections):
    """Calculate the average bbox size for debugging"""

    bbox = reflections["bbox"]
    sel = flex.random_selection(len(bbox), min(len(bbox), 1000))
    subset_bbox = bbox.select(sel)
    xmin, xmax, ymin, ymax, zmin, zmax = subset_bbox.parts()
    xsize = flex.mean((xmax - xmin).as_double())
    ysize = flex.mean((ymax - ymin).as_double())
    zsize = flex.mean((zmax - zmin).as_double())
    return xsize, ysize, zsize


@boost_adaptbx.boost.python.inject_into(Executor)
class _:
    @staticmethod
    def __getinitargs__():
        return ()


class _Job:
    def __init__(self):
        self.index = 0
        self.nthreads = 1


job = _Job()


class MultiProcessing:
    """
    Multi processing parameters
    """

    def __init__(self):
        self.method = "multiprocessing"
        self.nproc = 1
        self.njobs = 1
        self.nthreads = 1
        self.n_subset_split = None

    def update(self, other):
        self.method = other.method
        self.nproc = other.nproc
        self.njobs = other.njobs
        self.nthreads = other.nthreads
        self.n_subset_split = other.n_subset_split


class Lookup:
    """
    Lookup parameters
    """

    def __init__(self):
        self.mask = None

    def update(self, other):
        self.mask = other.mask


class Block:
    """
    Block parameters
    """

    def __init__(self):
        self.size = libtbx.Auto
        self.units = "degrees"
        self.threshold = 0.99
        self.force = False
        self.max_memory_usage = 0.90

    def update(self, other):
        self.size = other.size
        self.units = other.units
        self.threshold = other.threshold
        self.force = other.force
        self.max_memory_usage = other.max_memory_usage


class Shoebox:
    """
    Shoebox parameters
    """

    def __init__(self):
        self.flatten = False
        self.partials = False

    def update(self, other):
        self.flatten = other.flatten
        self.partials = other.partials


class Debug:
    """
    Debug parameters
    """

    def __init__(self):
        self.output = False
        self.select = None
        self.split_experiments = True
        self.separate_files = True

    def update(self, other):
        self.output = other.output
        self.select = other.select
        self.split_experiments = other.split_experiments
        self.separate_files = other.separate_files


class Parameters:
    """
    Class to handle parameters for the processor
    """

    def __init__(self):
        """
        Initialize the parameters
        """
        self.mp = MultiProcessing()
        self.lookup = Lookup()
        self.block = Block()
        self.shoebox = Shoebox()
        self.debug = Debug()

    def update(self, other):
        """
        Update the parameters
        """
        self.mp.update(other.mp)
        self.lookup.update(other.lookup)
        self.block.update(other.block)
        self.shoebox.update(other.shoebox)
        self.debug.update(other.debug)


def execute_parallel_task(task):
    """
    Helper function to run things on cluster
    """

    dials.util.log.config_simple_cached()
    result = task()
    handlers = logging.getLogger("dials").handlers
    assert len(handlers) == 1, "Invalid number of logging handlers"
    return result, handlers[0].records


class _Processor:
    """Processor interface class."""

    def __init__(self, manager):
        """
        Initialise the processor.

        The processor requires a manager class implementing the _Manager interface.
        This class executes all the workers in separate threads and accumulates the
        results to expose to the user.

        :param manager: The processing manager
        :param params: The phil parameters
        """
        self.manager = manager

    @property
    def executor(self):
        """
        Get the executor

        :return: The executor
        """
        return self.manager.executor

    @executor.setter
    def executor(self, function):
        """
        Set the executor

        :param function: The executor
        """
        self.manager.executor = function

    def process(self):
        """
        Do all the processing tasks.

        :return: The processing results
        """
        start_time = time()
        self.manager.initialize()
        mp_method = self.manager.params.mp.method
        mp_njobs = self.manager.params.mp.njobs
        mp_nproc = self.manager.params.mp.nproc

        assert mp_nproc > 0, "Invalid number of processors"
        if mp_nproc * mp_njobs > len(self.manager):
            mp_nproc = min(mp_nproc, len(self.manager))
            mp_njobs = int(math.ceil(len(self.manager) / mp_nproc))
        logger.info(self.manager.summary())
        if mp_njobs > 1:
            assert mp_method != "none" and mp_method is not None
            logger.info(
                " Using %s with %d parallel job(s) and %d processes per node\n",
                mp_method,
                mp_njobs,
                mp_nproc,
            )
        else:
            logger.info(" Using multiprocessing with %d parallel job(s)\n", mp_nproc)

        if mp_njobs * mp_nproc > 1:

            def process_output(result):
                rehandle_cached_records(result[1])
                self.manager.accumulate(result[0])

            multi_node_parallel_map(
                func=execute_parallel_task,
                iterable=list(self.manager.tasks()),
                njobs=mp_njobs,
                nproc=mp_nproc,
                callback=process_output,
                cluster_method=mp_method,
                preserve_order=True,
            )
        else:
            for task in self.manager.tasks():
                self.manager.accumulate(task())
        self.manager.finalize()
        end_time = time()
        self.manager.time.user_time = end_time - start_time
        result1, result2 = self.manager.result()
        return result1, result2, self.manager.time


class _ProcessorRot(_Processor):
    """Processor interface class for rotation data only."""

    def __init__(self, experiments, manager):
        """
        Initialise the processor.

        The processor requires a manager class implementing the _Manager interface.
        This class executes all the workers in separate threads and accumulates the
        results to expose to the user.

        :param manager: The processing manager
        """
        # Ensure we have the correct type of data
        if not experiments.all_sequences():
            raise RuntimeError(
                """
        An inappropriate processing algorithm may have been selected!
         Trying to perform rotation processing when not all experiments
         are indicated as rotation experiments.
      """
            )

        super().__init__(manager)


class NullTask:
    """
    A class to perform a null task.
    """

    def __init__(self, index, reflections):
        """
        Initialise the task

        :param index: The index of the processing job
        :param experiments: The list of experiments
        :param reflections: The list of reflections
        """
        self.index = index
        self.reflections = reflections

    def __call__(self):
        """
        Do the processing.

        :return: The processed data
        """
        return dials.algorithms.integration.Result(
            index=self.index,
            reflections=self.reflections,
            data=None,
            read_time=0,
            extract_time=0,
            process_time=0,
            total_time=0,
        )


class Task:
    """
    A class to perform a processing task.
    """

    def __init__(self, index, job, experiments, reflections, params, executor=None):
        """
        Initialise the task.

        :param index: The index of the processing job
        :param experiments: The list of experiments
        :param reflections: The list of reflections
        :param params: The processing parameters
        :param job: The frames to integrate
        :param flatten: Flatten the shoeboxes
        :param executor: The executor class
        """
        assert executor is not None, "No executor given"
        assert len(reflections) > 0, "Zero reflections given"
        self.index = index
        self.job = job
        self.experiments = experiments
        self.reflections = reflections
        self.params = params
        self.executor = executor

    def __call__(self):
        """
        Do the processing.

        :return: The processed data
        """
        # Get the start time
        start_time = time()

        # Set the global process ID
        job.index = self.index

        # Check all reflections have same imageset and get it
        exp_id = list(set(self.reflections["id"]))
        imageset = self.experiments[exp_id[0]].imageset
        for i in exp_id[1:]:
            assert (
                self.experiments[i].imageset == imageset
            ), "Task can only handle 1 imageset"

        # Get the sub imageset
        frame0, frame1 = self.job

        try:
            allowed_range = imageset.get_array_range()
        except Exception:
            allowed_range = 0, len(imageset)

        try:
            # range increasing
            assert frame0 < frame1

            # within an increasing range
            assert allowed_range[1] > allowed_range[0]

            # we are processing data which is within range
            assert frame0 >= allowed_range[0]
            assert frame1 <= allowed_range[1]

            # I am 99% sure this is implied by all the code above
            assert (frame1 - frame0) <= len(imageset)
            if len(imageset) > 1:
                imageset = imageset[frame0:frame1]
        except Exception as e:
            raise RuntimeError(f"Programmer Error: bad array range: {e}")

        try:
            frame0, frame1 = imageset.get_array_range()
        except Exception:
            frame0, frame1 = (0, len(imageset))

        self.executor.initialize(frame0, frame1, self.reflections)

        # Set the shoeboxes (don't allocate)
        self.reflections["shoebox"] = flex.shoebox(
            self.reflections["panel"],
            self.reflections["bbox"],
            allocate=False,
            flatten=self.params.shoebox.flatten,
        )

        # Create the processor
        processor = ShoeboxProcessor(
            self.reflections,
            len(imageset.get_detector()),
            frame0,
            frame1,
            self.params.debug.output,
        )

        # Loop through the imageset, extract pixels and process reflections
        read_time = 0.0
        for i in range(len(imageset)):
            st = time()
            image = imageset.get_corrected_data(i)
            if imageset.is_marked_for_rejection(i):
                mask = tuple(flex.bool(im.accessor(), False) for im in image)
            else:
                mask = imageset.get_mask(i)
                if self.params.lookup.mask is not None:
                    assert len(mask) == len(
                        self.params.lookup.mask
                    ), "Mask/Image are incorrect size %d %d" % (
                        len(mask),
                        len(self.params.lookup.mask),
                    )
                    mask = tuple(
                        m1 & m2 for m1, m2 in zip(self.params.lookup.mask, mask)
                    )

            read_time += time() - st
            processor.next(make_image(image, mask), self.executor)
            del image
            del mask
        assert processor.finished(), "Data processor is not finished"

        # Optionally save the shoeboxes
        if self.params.debug.output and self.params.debug.separate_files:
            output = self.reflections
            if self.params.debug.select is not None:
                output = output.select(self.params.debug.select(output))
            if self.params.debug.split_experiments:
                output = output.split_by_experiment_id()
                for table in output:
                    i = table["id"][0]
                    table.as_file("shoeboxes_%d_%d.refl" % (self.index, i))
            else:
                output.as_file("shoeboxes_%d.refl" % self.index)

        # Delete the shoeboxes
        if self.params.debug.separate_files or not self.params.debug.output:
            del self.reflections["shoebox"]

        # Finalize the executor
        self.executor.finalize()

        # Return the result
        return dials.algorithms.integration.Result(
            index=self.index,
            reflections=self.reflections,
            data=self.executor.data(),
            read_time=read_time,
            extract_time=processor.extract_time(),
            process_time=processor.process_time(),
            total_time=time() - start_time,
        )


class _Manager:
    """
    A class to manage processing book-keeping
    """

    def __init__(self, experiments, reflections, params):
        """
        Initialise the manager.

        :param experiments: The list of experiments
        :param reflections: The list of reflections
        :param params: The phil parameters
        """

        # Initialise the callbacks
        self.executor = None

        # Save some data
        self.experiments = experiments
        self.reflections = reflections

        # Other data
        self.data = {}

        # Save some parameters
        self.params = params

        # Set the finalized flag to False
        self.finalized = False

        # Initialise the timing information
        self.time = dials.algorithms.integration.TimingInfo()

    def initialize(self):
        """
        Initialise the processing
        """
        # Get the start time
        start_time = time()

        # Ensure the reflections contain bounding boxes
        assert "bbox" in self.reflections, "Reflections have no bbox"

        if self.params.mp.nproc is libtbx.Auto:
            self.params.mp.nproc = available_cores()
            logger.info(f"Setting nproc={self.params.mp.nproc}")

        # Compute the block size and processors
        self.compute_jobs()
        self.split_reflections()
        self.compute_processors()

        # Create the reflection manager
        self.manager = ReflectionManager(self.jobs, self.reflections)

        # Set the initialization time
        self.time.initialize = time() - start_time

    def task(self, index):
        """
        Get a task.
        """
        job = self.manager.job(index)
        frames = job.frames()
        expr_id = job.expr()
        assert expr_id[1] > expr_id[0], "Invalid experiment id"
        assert expr_id[0] >= 0, "Invalid experiment id"
        assert expr_id[1] <= len(self.experiments), "Invalid experiment id"
        experiments = self.experiments  # [expr_id[0]:expr_id[1]]
        reflections = self.manager.split(index)
        if len(reflections) == 0:
            logger.warning("No reflections in job %d ***", index)
            task = NullTask(index=index, reflections=reflections)
        else:
            task = Task(
                index=index,
                job=frames,
                experiments=experiments,
                reflections=reflections,
                params=self.params,
                executor=self.executor,
            )
        return task

    def tasks(self):
        """
        Iterate through the tasks.
        """
        for i in range(len(self)):
            yield self.task(i)

    def accumulate(self, result):
        """Accumulate the results."""
        self.data[result.index] = result.data
        self.manager.accumulate(result.index, result.reflections)
        self.time.read += result.read_time
        self.time.extract += result.extract_time
        self.time.process += result.process_time
        self.time.total += result.total_time

    def finalize(self):
        """
        Finalize the processing and finish.
        """
        # Get the start time
        start_time = time()

        # Check manager is finished
        assert self.manager.finished(), "Manager is not finished"

        # Update the time and finalized flag
        self.time.finalize = time() - start_time
        self.finalized = True

    def result(self):
        """
        Return the result.

        :return: The result
        """
        assert self.finalized, "Manager is not finalized"
        return self.manager.data(), self.data

    def finished(self):
        """
        Return if all tasks have finished.

        :return: True/False all tasks have finished
        """
        return self.finalized and self.manager.finished()

    def __len__(self):
        """
        Return the number of tasks.

        :return: the number of tasks
        """
        return len(self.manager)

    def compute_jobs(self):
        """
        Sets up a JobList() object in self.jobs
        """

        if self.params.block.size == libtbx.Auto:
            if (
                self.params.mp.nproc * self.params.mp.njobs == 1
                and not self.params.debug.output
                and not self.params.block.force
            ):
                self.params.block.size = None

        # calculate the block overlap based on the size of bboxes in the data
        # calculate once here rather than repeated in the loop below
        block_overlap = 0
        if self.params.block.size is not None:
            assert self.params.block.threshold > 0, "Threshold must be > 0"
            assert self.params.block.threshold <= 1.0, "Threshold must be < 1"
            frames_per_refl = sorted([b[5] - b[4] for b in self.reflections["bbox"]])
            cutoff = int(self.params.block.threshold * len(frames_per_refl))
            block_overlap = frames_per_refl[cutoff]

        groups = itertools.groupby(
            range(len(self.experiments)),
            lambda x: (id(self.experiments[x].imageset), id(self.experiments[x].scan)),
        )
        self.jobs = JobList()
        for key, indices in groups:
            indices = list(indices)
            i0 = indices[0]
            i1 = indices[-1] + 1
            expr = self.experiments[i0]
            scan = expr.scan
            imgs = expr.imageset
            array_range = (0, len(imgs))
            if scan is not None:
                assert len(imgs) >= len(scan), "Invalid scan range"
                array_range = scan.get_array_range()

            if self.params.block.size is None:
                block_size_frames = array_range[1] - array_range[0]
            elif self.params.block.size == libtbx.Auto:
                # auto determine based on nframes and overlap
                nframes = array_range[1] - array_range[0]
                nblocks = self.params.mp.nproc * self.params.mp.njobs
                # want data to be split into n blocks with overlaps
                # i.e. [x, overlap, y, overlap, y, overlap, ....,y,  overlap, x]
                # blocks are x + overlap, or overlap + y + overlap.
                x = (nframes - block_overlap) / nblocks
                block_size = int(math.ceil(x + block_overlap))
                # increase the block size to be at least twice the overlap, in
                # case the overlap is large e.g. if high mosaicity.
                block_size_frames = max(block_size, 2 * block_overlap)
            elif self.params.block.units == "radians":
                _, dphi = scan.get_oscillation(deg=False)
                block_size_frames = int(math.ceil(self.params.block.size / dphi))
                # if the specified block size is lower than the overlap,
                # reduce the overlap to be half of the block size.
                block_overlap = min(block_overlap, int(block_size_frames // 2))
            elif self.params.block.units == "degrees":
                _, dphi = scan.get_oscillation()
                block_size_frames = int(math.ceil(self.params.block.size / dphi))
                # if the specified block size is lower than the overlap,
                # reduce the overlap to be half of the block size.
                block_overlap = min(block_overlap, int(block_size_frames // 2))
            elif self.params.block.units == "frames":
                block_size_frames = int(math.ceil(self.params.block.size))
                block_overlap = min(block_overlap, int(block_size_frames // 2))
            else:
                raise RuntimeError(
                    f"Unknown block_size units {self.params.block.units!r}"
                )
            self.jobs.add(
                (i0, i1),
                array_range,
                block_size_frames,
                block_overlap,
            )
        assert len(self.jobs) > 0, "Invalid number of jobs"

    def split_reflections(self):
        """
        Split the reflections into partials or over job boundaries
        """

        # Optionally split the reflection table into partials, otherwise,
        # split over job boundaries
        if self.params.shoebox.partials:
            num_full = len(self.reflections)
            self.reflections.split_partials()
            num_partial = len(self.reflections)
            assert num_partial >= num_full, "Invalid number of partials"
            if num_partial > num_full:
                logger.info(
                    " Split %d reflections into %d partial reflections\n",
                    num_full,
                    num_partial,
                )
        else:
            num_full = len(self.reflections)
            self.jobs.split(self.reflections)
            num_partial = len(self.reflections)
            assert num_partial >= num_full, "Invalid number of partials"
            if num_partial > num_full:
                num_split = num_partial - num_full
                logger.info(
                    " Split %d reflections overlapping job boundaries\n", num_split
                )

        # Compute the partiality
        self.reflections.compute_partiality(self.experiments)

    def compute_processors(self):
        """
        Compute the number of processors
        """

        # Get the maximum shoebox memory to estimate memory use for one process
        memory_required_per_process = flex.max(
            self.jobs.shoebox_memory(self.reflections, self.params.shoebox.flatten)
        )

        (
            available_immediate_limit,
            available_incl_swap,
            report,
        ) = assess_available_memory(self.params)

        report.append(
            f"  {'Memory required per process':50}:{memory_required_per_process/1e9:5.1f} GB"
        )

        output_level = logging.INFO

        # Limit the number of parallel processes by amount of available memory
        if self.params.mp.method == "multiprocessing" and self.params.mp.nproc > 1:

            # Compute expected memory usage and warn if not enough
            njobs = available_immediate_limit / memory_required_per_process
            if njobs >= self.params.mp.nproc:
                # There is enough memory. Take no action
                pass
            elif njobs >= 1:
                # There is enough memory to run, but not as many processes as requested
                output_level = logging.WARNING
                report.append(
                    "Reducing number of processes from %d to %d due to memory constraints."
                    % (self.params.mp.nproc, int(njobs))
                )
                self.params.mp.nproc = int(njobs)
            elif (
                available_incl_swap * self.params.block.max_memory_usage
                >= memory_required_per_process
            ):
                # There is enough memory to run, but only if we count swap.
                output_level = logging.WARNING
                report.append(
                    "Reducing number of processes from %d to 1 due to memory constraints."
                    % self.params.mp.nproc,
                )
                report.append("Running this process will rely on swap usage!")
                self.params.mp.nproc = 1
            else:
                # There is not enough memory to run
                output_level = logging.ERROR

        report.append("")
        logger.log(output_level, "\n".join(report))

        if output_level >= logging.ERROR:
            raise MemoryError(
                """
          Not enough memory to run integration jobs.  This could be caused by a
          highly mosaic crystal model.  Possible solutions include increasing the
          percentage of memory allowed for shoeboxes or decreasing the block size.
          The average shoebox size is %d x %d pixels x %d images - is your crystal
          really this mosaic?
          """
                % _average_bbox_size(self.reflections)
            )

    def summary(self):
        """
        Get a summary of the processing
        """
        # Compute the task table
        if self.experiments.all_stills():
            rows = [["#", "Group", "Frame From", "Frame To", "# Reflections"]]
            for i in range(len(self)):
                job = self.manager.job(i)
                group = job.index()
                f0, f1 = job.frames()
                n = self.manager.num_reflections(i)
                rows.append([str(i), str(group), str(f0), str(f1), str(n)])
        elif self.experiments.all_sequences():
            rows = [
                [
                    "#",
                    "Group",
                    "Frame From",
                    "Frame To",
                    "Angle From",
                    "Angle To",
                    "# Reflections",
                ]
            ]
            for i in range(len(self)):
                job = self.manager.job(i)
                group = job.index()
                expr = job.expr()
                f0, f1 = job.frames()
                scan = self.experiments[expr[0]].scan
                p0 = scan.get_angle_from_array_index(f0)
                p1 = scan.get_angle_from_array_index(f1)
                n = self.manager.num_reflections(i)
                rows.append(
                    [str(i), str(group), str(f0 + 1), str(f1), str(p0), str(p1), str(n)]
                )
        else:
            raise RuntimeError("Experiments must be all sequences or all stills")

        # The job table
        task_table = tabulate(rows, headers="firstrow")

        # The format string
        if self.params.block.size is None:
            block_size = "auto"
        else:
            block_size = str(self.params.block.size)
        return (
            "Processing reflections in the following blocks of images:\n\n"
            " block_size: {} {}\n\n{}\n"
        ).format(
            block_size,
            "" if block_size in ("auto", "Auto") else self.params.block.units,
            task_table,
        )


class Processor3D(_ProcessorRot):
    """Top level processor for 3D processing."""

    def __init__(self, experiments, reflections, params):
        """Initialise the manager and the processor."""

        # Set some parameters
        params.shoebox.partials = False
        params.shoebox.flatten = False

        # Create the processing manager
        manager = _Manager(experiments, reflections, params)

        # Initialise the processor
        super().__init__(experiments, manager)


class ProcessorFlat3D(_ProcessorRot):
    """Top level processor for flat 3D processing."""

    def __init__(self, experiments, reflections, params):
        """Initialise the manager and the processor."""

        # Set some parameters
        params.shoebox.flatten = True
        params.shoebox.partials = False

        # Create the processing manager
        manager = _Manager(experiments, reflections, params)

        # Initialise the processor
        super().__init__(experiments, manager)


class Processor2D(_ProcessorRot):
    """Top level processor for 2D processing."""

    def __init__(self, experiments, reflections, params):
        """Initialise the manager and the processor."""

        # Set some parameters
        params.shoebox.partials = True

        # Create the processing manager
        manager = _Manager(experiments, reflections, params)

        # Initialise the processor
        super().__init__(experiments, manager)


class ProcessorSingle2D(_ProcessorRot):
    """Top level processor for still image processing."""

    def __init__(self, experiments, reflections, params):
        """Initialise the manager and the processor."""

        # Set some of the parameters
        params.block.size = 1
        params.block.units = "frames"
        params.shoebox.partials = True
        params.shoebox.flatten = False

        # Create the processing manager
        manager = _Manager(experiments, reflections, params)

        # Initialise the processor
        super().__init__(experiments, manager)


class ProcessorStills(_Processor):
    """Top level processor for still image processing."""

    def __init__(self, experiments, reflections, params):
        """Initialise the manager and the processor."""

        # Set some parameters
        params.block.size = 1
        params.block.units = "frames"
        params.shoebox.partials = False
        params.shoebox.flatten = False

        # Ensure we have the correct type of data
        if not experiments.all_stills():
            raise RuntimeError(
                """
        An inappropriate processing algorithm may have been selected!
         Trying to perform stills processing when not all experiments
         are indicated as stills experiments.
      """
            )

        # Create the processing manager
        manager = _Manager(experiments, reflections, params)

        # Initialise the processor
        super().__init__(manager)


def build_processor(Class, experiments, reflections, params=None):
    """
    A function to simplify building the processor

    :param Class: The input class
    :param experiments: The input experiments
    :param reflections: The reflections
    :param params: Optional input parameters
    """
    _params = Parameters()
    if params is not None:
        _params.update(params)

    return Class(experiments, reflections, _params)



"""
Weighting scheme definitions for scaling.
"""

from __future__ import annotations

from dials.array_family import flex


def get_weighting_scheme(Ih_table, weighting_scheme):
    """Return the appropriate weighting scheme from the params option."""
    if weighting_scheme == "invvar" or weighting_scheme is None:
        return WeightingBase(Ih_table)
    if weighting_scheme == "unity":
        return UnityWeights(Ih_table)
    if weighting_scheme == "GM":
        return GemanMcClureWeights(Ih_table)
    if weighting_scheme == "cauchy":
        return CauchyWeights(Ih_table)
    if weighting_scheme == "huber":
        return HuberWeights(Ih_table)
    else:
        raise ValueError(f"Invalid choice of weighting scheme: {weighting_scheme}")


class WeightingBase:
    """Base class that defines the properties of a Scaling Weights object."""

    weighting_scheme = "invvar"

    def __init__(self, Ih_table):
        self.Ih_table = Ih_table

    def calculate_initial_weights(self):
        """Set the initial weight, to allow first calculation of Ih."""
        self.Ih_table.weights = 1.0 / self.Ih_table.variances

    def apply_iterative_weights(self):
        """Update the weights, if relevant."""
        pass


class UnityWeights(WeightingBase):
    """Class for unity weights."""

    weighting_scheme = "unity"

    def calculate_initial_weights(self):
        self.Ih_table.weights = flex.double(self.Ih_table.size, 1.0)


class GemanMcClureWeights(WeightingBase):
    """Class for Geman-McClure weighting scheme (M-estimation)."""

    weighting_scheme = "GM"

    def apply_iterative_weights(self):
        e_i = (
            self.Ih_table.intensities
            - (self.Ih_table.inverse_scale_factors * self.Ih_table.Ih_values)
        ) / (self.Ih_table.inverse_scale_factors * self.Ih_table.Ih_values)
        self.Ih_table.weights = 1.0 / flex.pow2(1.0 + flex.pow2(e_i))


class CauchyWeights(WeightingBase):
    """Class for Geman-McClure weighting scheme (M-estimation)."""

    weighting_scheme = "cauchy"

    c = 2.385  # Value for 95% efficiency

    def apply_iterative_weights(self):
        e_i = (
            self.Ih_table.intensities
            - (self.Ih_table.inverse_scale_factors * self.Ih_table.Ih_values)
        ) / (self.Ih_table.inverse_scale_factors * self.Ih_table.Ih_values)
        self.Ih_table.weights = 1.0 / (1.0 + flex.pow2(e_i / self.c))


class HuberWeights(WeightingBase):
    """Class for Geman-McClure weighting scheme (M-estimation)."""

    weighting_scheme = "huber"

    c = 1.345  # Value for 95% efficiency

    def apply_iterative_weights(self):
        e_i = (
            self.Ih_table.intensities
            - (self.Ih_table.inverse_scale_factors * self.Ih_table.Ih_values)
        ) / (self.Ih_table.inverse_scale_factors * self.Ih_table.Ih_values)
        abs_ei = flex.abs(e_i)
        self.Ih_table.weights = flex.double(self.Ih_table.size, 1.0)
        sel = abs_ei > self.c
        sel_abs_ei = abs_ei.select(sel)
        sel_weight_fn = self.c / sel_abs_ei
        self.Ih_table.weights.set_selected(sel, sel_weight_fn)



# LIBTBX_SET_DISPATCHER_NAME cluster.dials.exec

from __future__ import annotations

import pickle

import dials.util


def get_cwd():
    """
    Get the current working directory
    """
    import sys

    return sys.argv[1]


def get_tid():
    """
    Get the task id
    """
    import os

    # FIXME At the moment, there is no portable way to know the task id through
    # drmaa. This is really annoying. So we have to use the SGE_TASK_ID.
    # Therefore, this will only work for SGE. We can probably add support for
    # other systems as and when needed.
    if "SGE_TASK_ID" in os.environ:
        return os.environ["SGE_TASK_ID"]
    else:
        raise KeyError("Could not find task id")


@dials.util.show_mail_handle_errors()
def run(_=None):
    import traceback
    from os.path import exists, join
    from time import sleep

    # Get the task id and the current working directory
    tid = get_tid()
    cwd = get_cwd()

    # Set the paths
    input_fn = join(cwd, f"{tid}.input")
    output_fn = join(cwd, f"{tid}.output")

    # Wait until it exists
    while not exists(input_fn):
        sleep(1)

    # Try to run the function, otherwise return an exception
    try:
        with open(input_fn, "rb") as infile:
            function, element = pickle.load(infile)

        result = function(element)
    except Exception as e:
        e.args = [traceback.format_exc()]
        result = e

    # Dump the result
    with open(output_fn, "wb") as outfile:
        pickle.dump(result, outfile, protocol=pickle.HIGHEST_PROTOCOL)


if __name__ == "__main__":
    run()



from __future__ import annotations

import numpy as np


class reader:
    """A class to read the CBF files used in DIALS"""

    def read_file(self, filename):
        """Read the CBF file"""
        import pycbf

        self.cbf_handle = pycbf.cbf_handle_struct()
        self.cbf_handle.read_file(filename, pycbf.MSG_DIGEST)
        self.cbf_handle.rewind_datablock()

    def get_data(self):
        """Get the gain array from the file"""
        from scitbx.array_family import flex

        # Select the first datablock and rewind all the categories
        self.cbf_handle.select_datablock(0)
        self.cbf_handle.select_category(0)
        self.cbf_handle.select_column(2)
        self.cbf_handle.select_row(0)

        # Check the type of the element to ensure it's a binary
        # otherwise raise an exception
        if "bnry" in self.cbf_handle.get_typeofvalue():
            # Read the image data into an array
            image_string = self.cbf_handle.get_integerarray_as_string()
            image = flex.int(np.fromstring(image_string, np.int32))

            # Get the array parameters
            parameters = self.cbf_handle.get_integerarrayparameters_wdims()
            image_size = (parameters[10], parameters[9])

            # Resize the image
            image.reshape(flex.grid(*image_size))
        else:
            raise TypeError("Can't find image")

        # Return the image
        return image


if __name__ == "__main__":
    import sys

    handle = reader()
    handle.read_file(sys.argv[1])
    image = handle.get_data()



from __future__ import annotations

import math

import scitbx.matrix
from dxtbx.model.detector_helpers import (
    get_detector_projection_2d_axes,
    get_panel_projection_2d_from_axes,
)
from scitbx.array_family import flex


def get_flex_image(
    data,
    vendortype,
    binning=1,
    brightness=1.0,
    saturation=65535.0,
    show_untrusted=False,
    color_scheme=0,
):
    # This is a combination of the get_data_type() and get_flex_image()
    # functions from iotbx.detectors.detectorbase.  XXX This may turn
    # out to be generally useful (see
    # e.g. rstbx.viewer.create_flex_image()), but where to place it?
    # dxtbx Format class?
    typehash = str(data.__class__)
    if typehash.find("int") >= 0:
        from iotbx.detectors import FlexImage
    elif typehash.find("double") >= 0:
        from iotbx.detectors import FlexImage_d as FlexImage

    return FlexImage(
        binning=binning,
        brightness=brightness,
        rawdata=data,
        saturation=int(round(saturation)),
        vendortype=vendortype,
        show_untrusted=show_untrusted,
        color_scheme=color_scheme,
    )


def get_flex_image_multipanel(
    detector,
    image_data,
    beam,
    brightness=1.0,
    binning=1,
    show_untrusted=False,
    color_scheme=0,
):
    # From xfel.cftbx.cspad_detector.readHeader() and
    # xfel.cftbx.cspad_detector.get_flex_image().  XXX Is it possible to
    # merge this with get_flex_image() above?  XXX Move to dxtbx Format
    # class (or a superclass for multipanel images)?

    from iotbx.detectors import generic_flex_image
    from libtbx.test_utils import approx_equal

    assert len(detector) == len(image_data), (len(detector), len(image_data))

    # Determine next multiple of eight of the largest panel size.
    data_max_focus = None
    for data in image_data:
        if data_max_focus is None:
            data_max_focus = data.focus()
        else:
            data_max_focus = (
                max(data_max_focus[0], data.focus()[0]),
                max(data_max_focus[1], data.focus()[1]),
            )
    data_padded = (
        8 * int(math.ceil(data_max_focus[0] / 8)),
        8 * int(math.ceil(data_max_focus[1] / 8)),
    )

    # Assert that all saturated values are equal and not None.  While
    # dxtbx records a separated trusted_range for each panel,
    # generic_flex_image supports only accepts a single common value for
    # the saturation.
    saturation = None
    for panel in detector:
        if saturation is None:
            saturation = panel.get_trusted_range()[1]
        else:
            assert approx_equal(saturation, panel.get_trusted_range()[1])
    assert saturation is not None

    # Create rawdata and flex_image_multipanel before populating it.
    rawdata = flex.double(flex.grid(len(detector) * data_padded[0], data_padded[1]))
    flex_image_multipanel = generic_flex_image(
        rawdata=rawdata,
        binning=binning,
        size1_readout=data_max_focus[0],
        size2_readout=data_max_focus[1],
        brightness=brightness,
        saturation=saturation,
        show_untrusted=show_untrusted,
        color_scheme=color_scheme,
    )

    # Calculate the average beam center across all panels, in meters
    # not sure this makes sense for detector which is not on a plane?
    beam_center = scitbx.matrix.col((0, 0, 0))
    npanels = 0
    for panel in detector:
        try:
            beam_center += scitbx.matrix.col(panel.get_beam_centre_lab(beam.get_s0()))
            npanels += 1
        except RuntimeError:  # catch DXTBX_ASSERT for no intersection
            pass
    beam_center /= npanels / 1e-3

    # XXX If a point is contained in two panels simultaneously, it will
    # be assigned to the panel defined first.  XXX Use a Z-buffer
    # instead?
    for i, panel in enumerate(detector):

        # Determine the pixel size for the panel (in meters), as pixel
        # sizes need not be identical.
        data = image_data[i]

        rawdata.matrix_paste_block_in_place(
            block=data.as_double(), i_row=i * data_padded[0], i_column=0
        )

        # If the panel already has a 2d projection then use it
        if panel.get_projection_2d():
            panel_r, panel_t = panel.get_projection_2d()
        else:
            if getattr(detector, "projection", "lab") == "image":
                if not hasattr(detector, "projection_2d_axes"):
                    # This can happen if the image wasn't preloaded as part
                    # of an experiment list Since it's used here, catch here.
                    detector.projection_2d_axes = get_detector_projection_2d_axes(
                        detector
                    )
                # Get axes from precalculated 2D projection.
                origin_2d, fast_2d, slow_2d = detector.projection_2d_axes
                fast = scitbx.matrix.col(fast_2d[i] + (0,))
                slow = scitbx.matrix.col(slow_2d[i] + (0,))
                origin = scitbx.matrix.col(origin_2d[i] + (0,)) * 1e-3
            else:
                # Get unit vectors in the fast and slow directions, as well as the
                # the locations of the origin and the center of the panel, in
                # meters. The origin is taken w.r.t. to average beam center of all
                # panels. This avoids excessive translations that can result from
                # rotations around the laboratory origin. Related to beam centre above
                # and dials#380 not sure this is right for detectors which are not
                # coplanar since system derived from first panel...
                fast = scitbx.matrix.col(panel.get_fast_axis())
                slow = scitbx.matrix.col(panel.get_slow_axis())
                origin = scitbx.matrix.col(panel.get_origin()) * 1e-3 - beam_center

            panel_r, panel_t = get_panel_projection_2d_from_axes(
                panel, data, fast, slow, origin
            )

        flex_image_multipanel.add_transformation_and_translation(panel_r, panel_t)

    flex_image_multipanel.followup_brightness_scale()
    return flex_image_multipanel



"""
Define an interface for observers and subjects.

A singleton decorator is also defined.
"""

from __future__ import annotations

import functools


def singleton(cls):
    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return getinstance


class Observer:
    def __init__(self):
        self.data = {}

    def update(self, subject):
        """Extract the relevant state information by inspecting the subject."""
        pass


class Subject:
    def __init__(self, events):
        self.observers = {}
        for event in events:
            self.observers[event] = {}

    @staticmethod
    def notify_event(event):
        """Define a decorator for notifying of a specific event."""

        def wrap(method):
            @functools.wraps(method)
            def notify(self, *args, **kwargs):
                r = method(self, *args, **kwargs)
                self.notify(event)
                return r

            return notify

        return wrap

    def get_observers(self, event):
        return self.observers[event]

    def register_observer(self, event, observer, callback=None):
        if callback is None:
            callback = getattr(observer, "update")
        else:
            callback = getattr(observer, callback)
        self.get_observers(event)[observer] = callback

    def unregister_observer(self, event, observer):
        del self.get_observers(event)[observer]

    def notify(self, event):
        for callback in self.get_observers(event).values():
            callback(self)



from __future__ import annotations

from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest

from cctbx.sgtbx import space_group, uctbx
from dxtbx import flumpy
from scitbx import sparse

from dials.algorithms.scaling.Ih_table import IhTable, IhTableBlock, map_indices_to_asu
from dials.array_family import flex


@pytest.fixture()
def large_reflection_table():
    """Create a reflection table."""
    return generate_refl_1()


@pytest.fixture()
def small_reflection_table():
    """Create a small reflection table."""
    return generate_refl_2()


@pytest.fixture(scope="module")
def test_sg():
    """Create a space group object."""
    return space_group("C 2y")


def mock_error_model():
    """Mock error model."""
    em = Mock()
    em.update_variances.return_value = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])
    return em


def refl_table_to_df(table):
    df = pd.DataFrame()
    for col in table.keys():
        if col != "asu_miller_index" and col != "miller_index":
            df[col] = flumpy.to_numpy(table[col])
    return df


def generate_refl_1():
    """Generate a test reflection table. Note that the variance values are chosen
    as the 'True' Ih_values, which would be found if unity weights were chosen
    in this example."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([100.0, 100.0, 80.0, 60.0, 30.0, 40.0, 60.0])
    reflections["inverse_scale_factor"] = flex.double(7, 2.0)
    reflections["variance"] = flex.double([90.0, 100.0, 90.0, 60.0, 30.0, 50.0, 50.0])
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (-1, 0, 0), (0, 2, 0), (0, 4, 0), (0, 0, -2), (0, 0, 2)]
    )
    reflections.set_flags(flex.bool(7, True), reflections.flags.integrated)
    return reflections


def generate_refl_2():
    """Generate another test reflection table."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([60.0, 30.0, 10.0, 30.0])
    reflections["variance"] = flex.double([60.0, 30.0, 10.0, 30.0])
    reflections["inverse_scale_factor"] = flex.double(4, 2.0)
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 4, 0), (10, 0, 0), (0, 4, 0)]
    )
    reflections.set_flags(flex.bool(4, True), reflections.flags.integrated)
    return reflections


def generated_refl_for_splitting_1():
    """ "Make a reflection table."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    reflections["variance"] = flex.double(6, 1.0)
    reflections["inverse_scale_factor"] = flex.double(6, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (2, 0, 0), (0, 0, 1), (2, 2, 2), (1, 0, 0), (2, 0, 0)]
    )
    reflections["d"] = flex.double([0.8, 2.1, 2.0, 1.4, 1.6, 2.5])
    reflections["partiality"] = flex.double(6, 1.0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 5.0),
            (0.0, 0.0, 8.0),
            (0.0, 0.0, 10.0),
            (0.0, 0.0, 12.0),
            (0.0, 0.0, 15.0),
        ]
    )
    reflections["s1"] = flex.vec3_double(
        [
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
        ]
    )
    reflections.set_flags(flex.bool(6, True), reflections.flags.integrated)
    reflections.set_flags(flex.bool(6, False), reflections.flags.bad_for_scaling)
    return reflections


def generated_refl_for_splitting_2():
    """Make a reflection table."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([7.0, 8.0, 9.0, 10.0, 11.0])
    reflections["variance"] = flex.double(5, 1.0)
    reflections["inverse_scale_factor"] = flex.double(5, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [(2, 2, 2), (2, 0, 0), (0, 0, 1), (2, 2, 2), (1, 0, 0)]
    )
    reflections["d"] = flex.double([0.8, 2.1, 2.0, 1.4, 1.6])
    reflections["partiality"] = flex.double(5, 1.0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 5.0),
            (0.0, 0.0, 8.0),
            (0.0, 0.0, 10.0),
            (0.0, 0.0, 12.0),
        ]
    )
    reflections["s1"] = flex.vec3_double(
        [
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
        ]
    )
    reflections.set_flags(flex.bool(5, True), reflections.flags.integrated)
    reflections.set_flags(flex.bool(5, False), reflections.flags.bad_for_scaling)
    return reflections


def test_IhTableblock_onedataset(large_reflection_table, test_sg):
    """Test direct initialisation of Ih_table block"""
    asu_indices = map_indices_to_asu(large_reflection_table["miller_index"], test_sg)
    large_reflection_table["asu_miller_index"] = asu_indices
    n_groups = len(set(asu_indices))
    n_refl = large_reflection_table.size()
    block = IhTableBlock(n_groups=n_groups, n_refl=n_refl)
    group_ids = np.array([0, 1, 0, 2, 3, 4, 4], dtype=np.uint64)

    df = refl_table_to_df(large_reflection_table)

    block.add_data(0, group_ids, df, large_reflection_table["miller_index"])

    # test properties
    assert block.inverse_scale_factors.size
    assert block.variances.size
    assert block.intensities.size
    assert block.weights.size
    assert block.asu_miller_index.size

    # assert list(block.block_selections[0]) == [0, 1, 2, 3, 4, 5, 6]
    assert len(block.block_selections) == 1

    assert block.h_index_matrix.n_rows == n_refl
    assert block.h_index_matrix.n_cols == n_groups
    assert block.h_index_matrix.non_zeroes == n_refl
    assert block.h_index_matrix[0, 0] == 1
    assert block.h_index_matrix[1, 1] == 1
    assert block.h_index_matrix[2, 0] == 1
    assert block.h_index_matrix[3, 2] == 1
    assert block.h_index_matrix[4, 3] == 1
    assert block.h_index_matrix[5, 4] == 1
    assert block.h_index_matrix[6, 4] == 1

    assert block.h_expand_matrix.n_cols == n_refl
    assert block.h_expand_matrix.n_rows == n_groups
    assert block.h_expand_matrix.non_zeroes == n_refl
    assert block.h_expand_matrix[0, 0] == 1
    assert block.h_expand_matrix[1, 1] == 1
    assert block.h_expand_matrix[0, 2] == 1
    assert block.h_expand_matrix[2, 3] == 1
    assert block.h_expand_matrix[3, 4] == 1
    assert block.h_expand_matrix[4, 5] == 1
    assert block.h_expand_matrix[4, 6] == 1

    assert list(block.intensities) == list(large_reflection_table["intensity"])
    assert list(block.weights) == list(1.0 / large_reflection_table["variance"])
    assert "Ih_values" not in block.Ih_table
    block.calc_Ih()
    assert list(block.Ih_values) == pytest.approx(
        [x / 2.0 for x in [90.0, 100.0, 90.0, 60.0, 30.0, 50.0, 50.0]]
    )

    with pytest.raises(AssertionError):
        block.add_data(
            0, group_ids, large_reflection_table, large_reflection_table["miller_index"]
        )

    assert list(block.calc_nh()) == [2, 1, 2, 1, 1, 2, 2]

    # Test update error model
    block.update_weights(mock_error_model())
    assert list(block.weights) == pytest.approx(
        [1.0, 1.0 / 2.0, 1.0 / 3.0, 1.0 / 4.0, 1.0 / 5.0, 1.0 / 6.0, 1.0 / 7.0]
    )

    # Test select method
    new_block = block.select(np.array([True, False, True, False, True, False, False]))

    assert list(new_block.block_selections[0]) == [0, 2, 4]

    assert new_block.h_index_matrix.n_rows == 3
    assert new_block.h_index_matrix.n_cols == 2
    assert new_block.h_index_matrix.non_zeroes == 3
    assert new_block.h_index_matrix[0, 0] == 1
    assert new_block.h_index_matrix[1, 0] == 1
    assert new_block.h_index_matrix[2, 1] == 1

    assert new_block.h_expand_matrix.n_cols == 3
    assert new_block.h_expand_matrix.n_rows == 2
    assert new_block.h_expand_matrix.non_zeroes == 3
    assert new_block.h_expand_matrix[0, 0] == 1
    assert new_block.h_expand_matrix[0, 1] == 1
    assert new_block.h_expand_matrix[1, 2] == 1

    assert list(new_block.Ih_values) == pytest.approx(
        [x / 2.0 for x in [90.0, 90.0, 30.0]]
    )
    with pd.option_context("mode.chained_assignment", None):
        # test setter methods
        new_block.inverse_scale_factors = np.array([1.0, 2.0, 3.0])
        assert list(new_block.inverse_scale_factors) == [1.0, 2.0, 3.0]
        with pytest.raises(AssertionError):
            new_block.inverse_scale_factors = np.array([1.0, 2.0, 3.0, 4.0])
        new_block.weights = np.array([0.1, 0.2, 0.3])
        assert list(new_block.weights) == [0.1, 0.2, 0.3]
        with pytest.raises(AssertionError):
            new_block.weights = np.array([1.0, 2.0, 3.0, 4.0])
        # test selecting initial table on groups, expecting same result as before
        new_block = block.select_on_groups(np.array([True, False, False, True, False]))
        assert list(new_block.block_selections[0]) == [0, 2, 4]

        assert new_block.h_index_matrix.n_rows == 3
        assert new_block.h_index_matrix.n_cols == 2
        assert new_block.h_index_matrix.non_zeroes == 3
        assert new_block.h_index_matrix[0, 0] == 1
        assert new_block.h_index_matrix[1, 0] == 1
        assert new_block.h_index_matrix[2, 1] == 1

        assert new_block.h_expand_matrix.n_cols == 3
        assert new_block.h_expand_matrix.n_rows == 2
        assert new_block.h_expand_matrix.non_zeroes == 3
        assert new_block.h_expand_matrix[0, 0] == 1
        assert new_block.h_expand_matrix[0, 1] == 1
        assert new_block.h_expand_matrix[1, 2] == 1

        assert list(new_block.Ih_values) == pytest.approx(
            [x / 2.0 for x in [90.0, 90.0, 30.0]]
        )


def test_IhTableblock_twodatasets(large_reflection_table, test_sg):
    """Test direct initialisation of Ih_table block. Use the same reflection
    table as previously to make comparisons easier"""
    asu_indices = map_indices_to_asu(large_reflection_table["miller_index"], test_sg)
    large_reflection_table["asu_miller_index"] = asu_indices
    n_groups = len(set(asu_indices))
    n_refl = large_reflection_table.size()

    sel_1 = flex.bool([True, True, False, False, True, True, True])
    dataset_1 = large_reflection_table.select(sel_1)
    dataset_2 = large_reflection_table.select(~sel_1)

    block = IhTableBlock(n_groups=n_groups, n_refl=n_refl, n_datasets=2)

    block.add_data(
        0,
        np.array([0, 1, 3, 4, 4], dtype=np.uint64),
        refl_table_to_df(dataset_1),
        dataset_1["miller_index"],
    )
    block.add_data(
        1,
        np.array([0, 2], dtype=np.uint64),
        refl_table_to_df(dataset_2),
        dataset_2["miller_index"],
    )

    assert list(block.block_selections[0]) == [0, 1, 2, 3, 4]
    assert list(block.block_selections[1]) == [0, 1]
    assert len(block.block_selections) == 2

    assert block.h_index_matrix.n_rows == n_refl
    assert block.h_index_matrix.n_cols == n_groups
    assert block.h_index_matrix.non_zeroes == n_refl
    # first dataset
    assert block.h_index_matrix[0, 0] == 1
    assert block.h_index_matrix[1, 1] == 1
    assert block.h_index_matrix[2, 3] == 1
    assert block.h_index_matrix[3, 4] == 1
    assert block.h_index_matrix[4, 4] == 1
    # then second dataset
    assert block.h_index_matrix[5, 0] == 1
    assert block.h_index_matrix[6, 2] == 1

    assert block.h_expand_matrix.n_cols == n_refl
    assert block.h_expand_matrix.n_rows == n_groups
    assert block.h_expand_matrix.non_zeroes == n_refl
    assert block.h_expand_matrix[0, 0] == 1
    assert block.h_expand_matrix[1, 1] == 1
    assert block.h_expand_matrix[3, 2] == 1
    assert block.h_expand_matrix[4, 3] == 1
    assert block.h_expand_matrix[4, 4] == 1
    assert block.h_expand_matrix[0, 5] == 1
    assert block.h_expand_matrix[2, 6] == 1

    # Test select method
    new_block = block.select(np.array([True, False, True, False, False, True, False]))

    assert list(new_block.block_selections[0]) == [0, 2]
    assert list(new_block.block_selections[1]) == [0]

    assert new_block.h_index_matrix.n_rows == 3
    assert new_block.h_index_matrix.n_cols == 2
    assert new_block.h_index_matrix.non_zeroes == 3
    assert new_block.h_index_matrix[0, 0] == 1
    assert new_block.h_index_matrix[1, 1] == 1
    assert new_block.h_index_matrix[2, 0] == 1

    assert new_block.h_expand_matrix.n_cols == 3
    assert new_block.h_expand_matrix.n_rows == 2
    assert new_block.h_expand_matrix.non_zeroes == 3
    assert new_block.h_expand_matrix[0, 0] == 1
    assert new_block.h_expand_matrix[1, 1] == 1
    assert new_block.h_expand_matrix[0, 2] == 1


def test_IhTable_split_into_blocks(
    large_reflection_table, small_reflection_table, test_sg
):
    """Test that the Ih_table datastructure correctly organises the data
    from two reflection tables into two IhTableBlocks."""

    sel1 = flex.bool(7, True)
    sel1[6] = False
    sel2 = flex.bool(4, True)
    sel2[1] = False

    Ih_table = IhTable(
        reflection_tables=[
            large_reflection_table.select(sel1),
            small_reflection_table.select(sel2),
        ],
        indices_lists=[sel1.iselection(), sel2.iselection()],
        space_group=test_sg,
        nblocks=2,
    )

    assert Ih_table.n_datasets == 2
    assert Ih_table.n_work_blocks == 2
    block_list = Ih_table.Ih_table_blocks
    assert list(block_list[0].asu_miller_index) == [
        (0, 0, 1),
        (0, 0, 2),
        (0, 2, 0),
    ]
    assert list(block_list[1].asu_miller_index) == [
        (0, 4, 0),
        (1, 0, 0),
        (1, 0, 0),
        (0, 4, 0),
        (1, 0, 0),
        (10, 0, 0),
    ]
    assert list(block_list[0].block_selections[0]) == [1, 5, 3]
    assert list(block_list[0].block_selections[1]) == []
    assert list(block_list[1].block_selections[0]) == [4, 0, 2]
    assert list(block_list[1].block_selections[1]) == [3, 0, 2]

    # test the 'get_block_selections_for_dataset' method
    block_sels_0 = Ih_table.get_block_selections_for_dataset(dataset=0)
    assert len(block_sels_0) == 2
    assert list(block_sels_0[0]) == [1, 5, 3]
    assert list(block_sels_0[1]) == [4, 0, 2]
    block_sels_1 = Ih_table.get_block_selections_for_dataset(dataset=1)
    assert len(block_sels_1) == 2
    assert list(block_sels_1[0]) == []
    assert list(block_sels_1[1]) == [3, 0, 2]

    # test the size method
    assert Ih_table.size == 9

    expected_h_idx_matrix = sparse.matrix(4, 3)
    expected_h_idx_matrix[0, 0] = 1
    expected_h_idx_matrix[1, 1] = 1
    expected_h_idx_matrix[2, 2] = 1
    assert block_list[0].h_index_matrix == expected_h_idx_matrix
    expected_h_idx_matrix = sparse.matrix(7, 3)
    expected_h_idx_matrix[0, 0] = 1
    expected_h_idx_matrix[1, 1] = 1
    expected_h_idx_matrix[2, 1] = 1
    expected_h_idx_matrix[3, 0] = 1
    expected_h_idx_matrix[4, 1] = 1
    expected_h_idx_matrix[5, 2] = 1
    assert block_list[1].h_index_matrix == expected_h_idx_matrix

    s = flex.double([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    assert list(block_list[1].sum_in_groups(s, "per_group")) == [5.0, 10.0, 6.0]
    assert list(block_list[1].sum_in_groups(s, "per_refl")) == [
        5.0,
        10.0,
        10.0,
        5.0,
        10.0,
        6.0,
    ]
    with pytest.raises(ValueError):
        block_list[1].sum_in_groups(s, "invalid")

    # check that only dataset 1 intensities are updated
    new_intensities = np.array([60.0, 50.0, 40.0, 30.0, 20.0, 10.0])
    Ih_table.update_data_in_blocks(
        data=new_intensities, dataset_id=0, column="intensity"
    )
    assert list(Ih_table.blocked_data_list[0].intensities) == [50.0, 10.0, 30.0]
    assert list(Ih_table.blocked_data_list[1].intensities) == [
        20.0,
        60.0,
        40.0,
        30.0,
        60.0,
        10.0,
    ]
    # try updating variances
    new_vars = np.array([100.0, 200.0, 300.0, 400.0])
    Ih_table.update_data_in_blocks(data=new_vars, dataset_id=1, column="variance")
    assert list(Ih_table.blocked_data_list[0].variances) == [100.0, 50.0, 60.0]
    assert list(Ih_table.blocked_data_list[1].variances) == [
        30.0,
        90.0,
        90.0,
        400.0,
        100.0,
        300.0,
    ]


def test_IhTable_freework(large_reflection_table, small_reflection_table, test_sg):
    sel1 = flex.bool(7, True)
    sel1[6] = False
    sel2 = flex.bool(4, True)
    sel2[1] = False

    Ih_table = IhTable(
        reflection_tables=[
            large_reflection_table.select(sel1),
            small_reflection_table.select(sel2),
        ],
        indices_lists=[sel1.iselection(), sel2.iselection()],
        space_group=test_sg,
        nblocks=2,
        free_set_percentage=50.0,
    )

    assert len(Ih_table.blocked_data_list) == 3
    assert Ih_table.n_datasets == 2
    assert Ih_table.n_work_blocks == 2
    block_list = Ih_table.Ih_table_blocks

    # two standard blocks
    assert block_list[0].h_index_matrix[0, 0] == 1
    assert block_list[0].h_index_matrix.non_zeroes == 1
    assert block_list[1].h_index_matrix[0, 0] == 1
    assert block_list[1].h_index_matrix[1, 0] == 1
    assert block_list[1].h_index_matrix[2, 1] == 1
    assert block_list[1].h_index_matrix.non_zeroes == 3
    # free set block
    assert block_list[2].h_index_matrix[0, 0] == 1
    assert block_list[2].h_index_matrix[1, 1] == 1
    assert block_list[2].h_index_matrix[2, 2] == 1
    assert block_list[2].h_index_matrix[3, 2] == 1
    assert block_list[2].h_index_matrix[4, 2] == 1
    assert block_list[2].h_index_matrix.non_zeroes == 5

    assert list(block_list[0].block_selections[0]) == [5]
    assert list(block_list[0].block_selections[1]) == []
    assert list(block_list[1].block_selections[0]) == [4]
    assert list(block_list[1].block_selections[1]) == [3, 2]
    assert list(block_list[2].block_selections[0]) == [1, 3, 0, 2]
    assert list(block_list[2].block_selections[1]) == [0]

    # test get_block_selections_for_dataset
    block_sels_0 = Ih_table.get_block_selections_for_dataset(0)
    assert len(block_sels_0) == 3
    assert list(block_sels_0[0]) == [5]
    assert list(block_sels_0[1]) == [4]
    assert list(block_sels_0[2]) == [1, 3, 0, 2]
    block_sels_1 = Ih_table.get_block_selections_for_dataset(1)
    assert len(block_sels_1) == 3
    assert list(block_sels_1[0]) == []
    assert list(block_sels_1[1]) == [3, 2]
    assert list(block_sels_1[2]) == [0]
    with pytest.raises(AssertionError):
        _ = Ih_table.get_block_selections_for_dataset(2)

    Ih_table.calc_Ih()

    # test setting data
    # set scale factors
    new_s_block_2 = np.arange(start=1, stop=6)
    Ih_table.set_inverse_scale_factors(new_s_block_2, 2)
    assert list(Ih_table.Ih_table_blocks[2].inverse_scale_factors) == list(
        new_s_block_2
    )
    # set derivatives
    derivs = Mock()
    Ih_table.set_derivatives(derivs, 0)
    assert Ih_table.Ih_table_blocks[0].derivatives is derivs

    def update_vars_side_effect(*args):
        return np.array([0.5] * len(args[0]))

    # test setting an error model
    em = Mock()
    em.update_variances.side_effect = update_vars_side_effect

    Ih_table.update_weights(em)
    for block in Ih_table.Ih_table_blocks:
        assert list(block.weights) == pytest.approx([2.0] * block.size)

    Ih_table.calc_Ih(1)

    # now test free set with offset
    Ih_table = IhTable(
        reflection_tables=[
            large_reflection_table.select(sel1),
            small_reflection_table.select(sel2),
        ],
        indices_lists=[sel1.iselection(), sel2.iselection()],
        space_group=test_sg,
        nblocks=2,
        free_set_percentage=50.0,
        free_set_offset=1,
    )
    assert len(Ih_table.blocked_data_list) == 3
    assert Ih_table.n_datasets == 2
    assert Ih_table.n_work_blocks == 2
    block_list = Ih_table.Ih_table_blocks

    # two standard blocks
    assert block_list[0].h_index_matrix[0, 0] == 1
    assert block_list[0].h_index_matrix[1, 1] == 1
    assert block_list[0].h_index_matrix.non_zeroes == 2
    assert block_list[1].h_index_matrix[0, 0] == 1
    assert block_list[1].h_index_matrix[1, 0] == 1
    assert block_list[1].h_index_matrix[2, 0] == 1
    assert block_list[1].h_index_matrix.non_zeroes == 3
    # free set block
    assert block_list[2].h_index_matrix[0, 0] == 1
    assert block_list[2].h_index_matrix[1, 1] == 1
    assert block_list[2].h_index_matrix[2, 1] == 1
    assert block_list[2].h_index_matrix[3, 2] == 1
    assert block_list[2].h_index_matrix.non_zeroes == 4

    assert list(block_list[0].block_selections[0]) == [1, 3]
    assert list(block_list[0].block_selections[1]) == []
    assert list(block_list[1].block_selections[0]) == [0, 2]
    assert list(block_list[1].block_selections[1]) == [0]
    assert list(block_list[2].block_selections[0]) == [5, 4]
    assert list(block_list[2].block_selections[1]) == [3, 2]

    Ih_table.calc_Ih()

    # Test the 'as_miller_array' method.
    unit_cell = uctbx.unit_cell((1.0, 1.0, 1.0, 90.0, 90.0, 90.0))

    arr = Ih_table.as_miller_array(unit_cell)
    assert arr.size() == 5
    assert list(arr.indices()) == [
        (0, 0, 1),
        (0, 2, 0),
        (1, 0, 0),
        (1, 0, 0),
        (1, 0, 0),
    ]

    arr = Ih_table.as_miller_array(unit_cell, return_free_set_data=True)
    assert arr.size() == 4
    assert list(arr.indices()) == [(0, 0, 2), (0, 4, 0), (0, 4, 0), (10, 0, 0)]


def test_set_Ih_values_to_target(test_sg):
    """Test the setting of Ih values for targeted scaling."""
    """Generate input for testing joint_Ih_table."""
    Ih_table = IhTable(
        [generated_refl_for_splitting_1(), generated_refl_for_splitting_2()],
        test_sg,
        nblocks=2,
    )
    Ih_table.calc_Ih()
    # First check that values are set up correctly.
    block_list = Ih_table.blocked_data_list
    assert list(block_list[0].Ih_values) == pytest.approx(
        [6.0, 17.0 / 3.0, 17.0 / 3.0, 6.0, 17.0 / 3.0]
    )
    assert list(block_list[1].Ih_values) == pytest.approx(
        [16.0 / 3.0, 16.0 / 3.0, 7.0, 16.0 / 3.0, 7.0, 7.0]
    )

    # set some values in the target
    # change the (2, 0, 0) reflections to (4, 0, 0) to test if they are removed
    # from the blocks
    t1 = generated_refl_for_splitting_1()
    t2 = generated_refl_for_splitting_2()
    t1["miller_index"][1] = (4, 0, 0)
    t1["miller_index"][5] = (4, 0, 0)
    t2["miller_index"][1] = (4, 0, 0)
    target = IhTable([t1, t2], test_sg, nblocks=1)
    target.blocked_data_list[0].Ih_table["Ih_values"] = np.array(
        [0.1, 0.2, 0.2, 0.3, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4, 0.4]
    )

    for block in block_list:
        block.match_Ih_values_to_target(target)
    assert list(block_list[0].Ih_values) == [0.1, 0.2, 0.2, 0.1, 0.2]
    assert list(block_list[1].Ih_values) == [0.4, 0.4, 0.4]


'''@pytest.mark.xfail(reason='not yet updated code')
def test_apply_iterative_weighting(reflection_table_for_block, test_sg):
  """Test the setting of iterative weights."""

  Ihtableblock = IhTableBlock(reflection_table_for_block[0], test_sg,
    weighting_scheme='GM')

  # After construction, weights should initially be set to inverse variances,
  # to allow the first calculation of Ih (by least-squares approach).
  assert list(Ihtableblock.weights) == [1.0] * 6

  # Now update weights
  Ihtableblock.update_weights()
  assert list(Ihtableblock.weights) != [1.0] * 6 # Check changed, or that
  # new weights are not by chance identical to old weights.
  gIh = Ihtableblock.inverse_scale_factors * Ihtableblock.Ih_values
  t = (Ihtableblock.intensities - gIh) / gIh
  assert list(Ihtableblock.weights) == list(1.0/(1.0 + t**2)**2)


  Ih_table = IhTable([reflection_table_for_block[0]], test_sg,
    weighting_scheme='GM')
  block = Ih_table.blocked_data_list[0]
  assert list(block.weights) == [1.0] * 6
  Ih_table.update_weights()
  gIh = block.inverse_scale_factors * block.Ih_values
  t = (block.intensities - gIh) / gIh
  assert list(block.weights) == list(1.0/(1.0 + t**2)**2)'''



from __future__ import annotations

from dxtbx.model.experiment_list import ExperimentListFactory
from scitbx import matrix

from dials.algorithms.integration import Corrections, CorrectionsMulti
from dials.array_family import flex


def test_run(dials_data):
    filename = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    exlist = ExperimentListFactory.from_json_file(filename)
    assert len(exlist) == 1

    rlist = flex.reflection_table.from_predictions_multi(exlist)

    corrector = CorrectionsMulti()
    for experiment in exlist:
        corrector.append(
            Corrections(experiment.beam, experiment.goniometer, experiment.detector)
        )

    lp1 = corrector.lp(rlist["id"], rlist["s1"])

    lp2 = flex.double(
        [LP_calculations(exlist[i], s1) for i, s1 in zip(rlist["id"], rlist["s1"])]
    )

    diff = flex.abs(lp1 - lp2)
    assert diff.all_lt(1e-7)


def LP_calculations(experiment, s1):
    """See Kabsch, J. Appl. Cryst 1988 21 916-924."""

    tpl_n = experiment.beam.get_polarization_normal()
    tpl_s0 = experiment.beam.get_s0()
    tpl_m2 = experiment.goniometer.get_rotation_axis()
    tpl_s1 = s1
    p = experiment.beam.get_polarization_fraction()

    n = matrix.col(tpl_n)
    s0 = matrix.col(tpl_s0)
    u = matrix.col(tpl_m2)
    s = matrix.col(tpl_s1)

    L_f = abs(s.dot(u.cross(s0))) / (s.length() * s0.length())

    P_f = (1 - 2 * p) * (1 - (n.dot(s) / s.length()) ** 2.0) + p * (
        1 + (s.dot(s0) / (s.length() * s0.length())) ** 2.0
    )

    return L_f / P_f



from __future__ import annotations

import copy

import numpy as np

from dxtbx import flumpy
from dxtbx.model import ExperimentList
from scitbx.array_family import flex

from dials.algorithms.integration.ssx.ssx_integrate import (
    OutputAggregator,
    OutputCollector,
    SimpleIntegrator,
)
from dials.algorithms.profile_model.ellipsoid.algorithm import (
    final_integrator,
    initial_integrator,
    predict,
    predict_after_ellipsoid_refinement,
    run_ellipsoid_refinement,
)
from dials.algorithms.profile_model.ellipsoid.indexer import reindex
from dials.algorithms.profile_model.ellipsoid.refiner import (
    BadSpotForIntegrationException,
)
from dials.constants import FULL_PARTIALITY


class ToFewReflections(Exception):
    pass


class EllipsoidOutputCollector(OutputCollector):
    def collect_after_refinement(self, experiment, reflection_table, refiner_output):
        if "initial_rmsd_x" not in self.data:
            self.data["initial_rmsd_x"] = refiner_output[0][0]["rmsd"][0]
            self.data["initial_rmsd_y"] = refiner_output[0][0]["rmsd"][1]
        self.data["final_rmsd_x"] = refiner_output[0][-1]["rmsd"][0]
        self.data["final_rmsd_y"] = refiner_output[0][-1]["rmsd"][1]
        # also record some model info:
        self.data["profile_model"] = experiment.profile.to_dict()
        self.data["profile_model_mosaicity"] = experiment.profile.mosaicity()

    def collect_after_preprocess(self, experiment, reflection_table):
        self.data["n_strong_after_preprocess"] = reflection_table.size()
        sel = reflection_table.get_flags(reflection_table.flags.integrated_sum)
        n_sum = sel.count(True)
        self.data["n_strong_sum_integrated"] = n_sum
        if n_sum:
            xobs, yobs, _ = reflection_table["xyzobs.px.value"].select(sel).parts()
            xcal, ycal, _ = reflection_table["xyzcal.px"].select(sel).parts()
            rmsd = flex.mean((xobs - xcal) ** 2 + (yobs - ycal) ** 2) ** 0.5
            self.data["strong_rmsd_preprocessed"] = rmsd
        else:
            self.data["strong_rmsd_preprocessed"] = 0.0

    def collect_after_integration(self, experiment, reflection_table):
        super().collect_after_integration(experiment, reflection_table)
        p = flumpy.to_numpy(reflection_table["partiality"])
        bins = np.linspace(0, 1, 11)
        hist = np.histogram(p, bins)
        self.data["partiality"] = hist[0]


class EllipsoidIntegrator(SimpleIntegrator):
    def __init__(self, params, collect_data=False):
        super().__init__(params)
        if collect_data:
            self.collector = EllipsoidOutputCollector()

    def run(self, experiment, table):

        # first set ids to zero so can integrate (this is how integration
        # finds the image in the imageset)
        ids_map = dict(table.experiment_identifiers())
        table["id"] = flex.int(table.size(), 0)
        del table.experiment_identifiers()[list(ids_map.keys())[0]]
        table.experiment_identifiers()[0] = list(ids_map.values())[0]

        fix_list = []
        if self.params.profile.ellipsoid.unit_cell.fixed:
            fix_list.append("unit_cell")
        if self.params.profile.ellipsoid.orientation.fixed:
            fix_list.append("orientation")

        initial_table = copy.deepcopy(table)

        self.collector.initial_collect(experiment, initial_table)

        for i in range(self.params.profile.ellipsoid.refinement.n_macro_cycles):
            try:
                if i == 0:
                    table, sigma_d = self.preprocess(
                        experiment, initial_table, self.params
                    )
                    self.collector.collect_after_preprocess(experiment, table)
                else:
                    # update the predictions and s1 vectors using the latest models
                    initial_table = predict_after_ellipsoid_refinement(
                        experiment, initial_table
                    )
                    table, sigma_d = self.preprocess(
                        experiment, initial_table, self.params
                    )
            except ToFewReflections as e:
                raise RuntimeError(e)
            else:
                try:
                    experiment, table, refiner_output = self.refine(
                        experiment,
                        table,
                        sigma_d,
                        profile_model=self.params.profile.ellipsoid.rlp_mosaicity.model,
                        fix_list=fix_list,
                        n_cycles=self.params.profile.ellipsoid.refinement.n_cycles,
                        capture_progress=isinstance(
                            self.collector, EllipsoidOutputCollector
                        ),
                    )
                except BadSpotForIntegrationException as e:
                    raise RuntimeError(e)
                else:
                    self.collector.collect_after_refinement(
                        experiment, table, refiner_output["refiner_output"]["history"]
                    )

        predicted = self.predict(
            experiment,
            table,
            d_min=self.params.prediction.d_min,
            prediction_probability=self.params.profile.ellipsoid.prediction.probability,
        )
        # do we want to add unmatched i.e. strong spots which weren't predicted?
        self.collector.collect_after_prediction(predicted, table)

        predicted = self.integrate(experiment, predicted, sigma_d)
        self.collector.collect_after_integration(experiment, predicted)

        return experiment, predicted, self.collector

    # methods below are staticmethods, so that one can build different
    # workflows from the run method above using individual algorithm components

    @staticmethod
    def preprocess(experiment, reflection_table, params):
        reference = reindex(
            reflection_table,
            experiment,
            outlier_probability=params.profile.ellipsoid.refinement.outlier_probability,
            max_separation=params.profile.ellipsoid.refinement.max_separation,
            fail_on_bad_index=params.profile.ellipsoid.indexing.fail_on_bad_index,
        )
        if reference.size() < params.profile.ellipsoid.refinement.min_n_reflections:
            raise ToFewReflections(
                "Too few reflections to perform refinement: got %d, expected %d"
                % (
                    reference.size(),
                    params.profile.ellipsoid.refinement.min_n_reflections,
                )
            )
        reference, sigma_d = initial_integrator(ExperimentList([experiment]), reference)

        return reference, sigma_d

    @staticmethod
    def refine(
        experiment,
        reflection_table,
        sigma_d,
        profile_model,
        fix_list=None,
        n_cycles=1,
        capture_progress=False,
    ):
        fix_unit_cell = False
        fix_orientation = False
        if fix_list:
            if "unit_cell" in fix_list:
                fix_unit_cell = True
            if "orientation" in fix_list:
                fix_orientation = True

        expts, refls, output_data = run_ellipsoid_refinement(
            ExperimentList([experiment]),
            reflection_table,
            sigma_d,
            profile_model=profile_model,
            fix_unit_cell=fix_unit_cell,
            fix_orientation=fix_orientation,
            n_cycles=n_cycles,
            capture_progress=capture_progress,
        )
        return expts[0], refls, output_data

    @staticmethod
    def predict(
        experiment, reference, d_min=None, prediction_probability=FULL_PARTIALITY
    ):
        id_map = dict(reference.experiment_identifiers())
        reflection_table = predict(
            ExperimentList([experiment]),
            d_min=d_min,
            prediction_probability=prediction_probability,
        )
        ids_ = set(reflection_table["id"])
        assert ids_ == set(id_map.keys()), f"{ids_}, {id_map.keys()}"
        for id_ in ids_:
            reflection_table.experiment_identifiers()[id_] = id_map[id_]
        return reflection_table

    @staticmethod
    def integrate(experiment, reflection_table, sigma_d):
        reflection_table = final_integrator(
            ExperimentList([experiment]),
            reflection_table,
            sigma_d,
        )
        return reflection_table


class EllipsoidOutputAggregator(OutputAggregator):
    def make_plots(self):
        plots = super().make_plots()
        initial_rmsds_x = [d["initial_rmsd_x"] for d in self.data.values()]
        final_rmsds_x = [d["final_rmsd_x"] for d in self.data.values()]
        initial_rmsds_y = [d["initial_rmsd_y"] for d in self.data.values()]
        final_rmsds_y = [d["final_rmsd_y"] for d in self.data.values()]
        n = list(self.data.keys())
        hist = np.zeros((10,))
        for d in self.data.values():
            hist += d["partiality"]
        bins = np.linspace(0.05, 0.95, 10)
        rmsd_plots = {
            "refinement_rmsds": {
                "data": [
                    {
                        "x": n,
                        "y": initial_rmsds_x,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Initial rmsd_x",
                    },
                    {
                        "x": n,
                        "y": final_rmsds_x,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Final rmsd_x",
                    },
                    {
                        "x": n,
                        "y": initial_rmsds_y,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Initial rmsd_y",
                    },
                    {
                        "x": n,
                        "y": final_rmsds_y,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Final rmsd_y",
                    },
                ],
                "layout": {
                    "title": "Rmsds of integrated reflections per image",
                    "xaxis": {"title": "image number"},
                    "yaxis": {"title": "RMSD (px)"},
                },
            },
            "partiality": {
                "data": [
                    (
                        {
                            "x": list(bins),
                            "y": list(hist),
                            "type": "scatter",
                            "mode": "markers",
                        }
                    )
                ],
                "layout": {
                    "title": "Partiality distribution",
                    "xaxis": {"title": "Partiality bin centre"},
                    "yaxis": {"title": "Number of reflections"},
                },
            },
        }
        plots.update(rmsd_plots)
        return plots
